//	KaleidoTile-Common.h
//
//	Platform-independent definitions for KaleidoTile's internal code.
//	The internal code doesn't know or care what platform (Win32 or MacOS)
//	it's running on.
//	However, the platform-specific user interface *does* include
//	the present file.  For that reason we take care not to use any
//	OpenGL-specific definitions (i.e. nothing from gl.h) because
//	the user interface doesn't know about OpenGL.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#pragma once

#include <stdlib.h>	//	for malloc() and free()
#include <stddef.h>	//	for offsetof()
#include "KaleidoTileMatrices.h"
#include "KaleidoTileTiling.h"
#include "KaleidoTileVelocities.h"
#include "Matrix44.h"

//	Acknowledge unused parameters.
#define UNUSED_PARAMETER(x)	((void)(x))

//	Define a byte.
typedef unsigned char	Byte;

//	Internal functions pass error messages as pointers
//	to zero-terminated UTF-8 strings.  The strings are defined
//	statically, so the caller need not (must not!) free them.
typedef const char		*ErrorText;

//	Define GET_MEMORY and FREE_MEMORY macros to manage memory so
//	that in future we can easily change memory allocation schemes.
//	Use a variable gMemCount to test for memory leaks.
//	Define FREE_MEMORY_SAFELY() to test for and set NULL pointers.
#define GET_MEMORY(n)			(gMemCount++, malloc(n))
#define FREE_MEMORY(p)			(gMemCount--, free(p))
#define FREE_MEMORY_SAFELY(p)	{if (p != NULL) {FREE_MEMORY(p); p = NULL;}}

//	BUFFER_LENGTH() measure the number of items in an array,
//	not the number of bytes, and automatically adjusts to changes
//	in the number of elements or the size of each element.
#define BUFFER_LENGTH(a)		( sizeof(a) / sizeof((a)[0]) )

//	Maximum and minimum macros.
//	Call them MAXIMUM() and MINIMUM() to avoid conflicts with Cocoa's MAX() and MIN().
#define MAXIMUM(A,B)			(((A) >= (B)) ? (A) : (B))
#define MINIMUM(A,B)			(((A) <= (B)) ? (A) : (B))

//	Macro to compare two 2-letter language codes.
#define SAME_LANGUAGE_CODE(A,B)	((A)[0] == (B)[0] && (A)[1] == (B)[1])

//	RUNNING_OPENGL_3 marks places where changes will be required
//	for an OpenGL 3.0 forward-compatible context.
//	When this program supports such a context,
//	the fixed macro RUNNING_OPENGL_3 will get replaced
//	by a flag saying whether or not an OpenGL 3.0 context
//	has been loaded at runtime.  It's possible that more than
//	one such flag will be required, depending on how closely
//	OpenGL ES 2.0 mirrors OpenGL 3.0.  For now, this macro
//	suffices to mark places that will require attention.
#define RUNNING_OPENGL_3	false

//	Compile a version with custom options for use in talks.
//#define FOR_TALKS


//	Represent colors as (αR, αG, αB, α) instead of the traditional (R,G,B,α)
//
//		1.	to allow rigorously correct blending
//			[see comments preceding the call to glBlendFunc()], and
//
//		2.	to let mipmap generation automatically average pixels
//			weighted by their alpha values [see comments preceding
//			the call to glGenerateMipmap()].
//
#define PREMULTIPLY_RGBA(r,g,b,a)	(float [4]){(a)*(r), (a)*(g), (a)*(b), (a)}


//	Assign the vertex attributes to well-defined locations in the vertex shader.
//	ATTRIBUTE_POSITION must bind to location 0, but the remaining choices are arbitrary.
#define ATTRIBUTE_POSITION				0
#define ATTRIBUTE_MVP_MATRIX_ROW_0		1
#define ATTRIBUTE_MVP_MATRIX_ROW_1		2
#define ATTRIBUTE_MVP_MATRIX_ROW_2		3
#define ATTRIBUTE_MVP_MATRIX_ROW_3		4
#define ATTRIBUTE_DIFFUSE_EVALUATOR		5
#define ATTRIBUTE_SPECULAR_EVALUATOR	6
#define ATTRIBUTE_NORMAL				7
#define ATTRIBUTE_TEX_COORD				8
#define ATTRIBUTE_COLOR_HSLA			9

//	Keep an array of shader programs, each referenced by an unsigned integer 
//	that glCreateProgram() provides to refer to the given program.
//	Each program contains a vertex shader and a fragment shader.
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
typedef enum
{
	ShaderColor = 0,
	ShaderTexture,
	ShaderColorWithTexture,
	ShaderColorWithAlphaMask,
	ShaderColorWithLighting,
	ShaderTextureWithLighting,
	ShaderHSLAWithTexture,
	NumShaders
} ShaderIndex;

//	Keep an array of texture "names", each "name" being an unsigned integer 
//	that glGenTextures() generates to refer to the given texture.  
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
typedef enum
{
	TextureFaceBase	= 0,
	TextureFaceA			= TextureFaceBase + 0,
	TextureFaceB			= TextureFaceBase + 1,
	TextureFaceC			= TextureFaceBase + 2,
	TextureTilingBackground	= TextureFaceBase + 3,
	TextureRegularName,
	
	TextureControlBackground,
	TextureLabelBase,
	TextureLabelSymmetry		= TextureLabelBase + 0,
	TextureLabelStyle			= TextureLabelBase + 1,
	TextureLabelControlPoint	= TextureLabelBase + 2,
	TextureLabelFaces			= TextureLabelBase + 3,
	TextureLabelBackground		= TextureLabelBase + 4,
	
	TextureSymmetryBase,
	TextureSymmetry3	= TextureSymmetryBase + 0,
	TextureSymmetry4	= TextureSymmetryBase + 1,
	TextureSymmetry5	= TextureSymmetryBase + 2,
	TextureSymmetry6	= TextureSymmetryBase + 3,
	TextureSymmetry7	= TextureSymmetryBase + 4,
	
	TextureStyleBase,
	TextureStyleFlat	= TextureStyleBase + 0,
	TextureStyleCurved	= TextureStyleBase + 1,
	
	TextureRoundMarker,
	
	TextureBevel,
	TextureCheckerboard,
	TextureSquareMarker,
	TextureFacePalette,
	TextureBackgroundPalette,
	TextureButtonBase,
	TextureButtonHueSaturation	= TextureButtonBase + 0,
	TextureButtonLightnessAlpha	= TextureButtonBase + 1,
	TextureButtonTexture		= TextureButtonBase + 2,
	TextureButtonPaste			= TextureButtonBase + 3,
	TextureButtonNull			= TextureButtonBase + 4,
	
	NumTextures
} TextureIndex;

//	Keep an array of vertex buffer "names" and a parallel array 
//	of so-called index buffer "names", each "name" being an unsigned integer
//	that glGenBuffers() generates to refer to the given vertex or index buffer.  
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.  
typedef enum
{
		//	VertexBufferMeshA, VertexBufferMeshB and VertexBufferMeshC
		//	index the mesh names for the three parts of the fundamental triangle.
	VertexBufferMeshBase = 0,
	VertexBufferMeshA = VertexBufferMeshBase + 0,
	VertexBufferMeshB = VertexBufferMeshBase + 1,
	VertexBufferMeshC = VertexBufferMeshBase + 2,
	NumVertexBuffers
} VertexBufferIndex;

//	Stereoscopic 3D modes
typedef enum
{
	StereoNone,
	StereoAnaglyphic,	//	superimposed red and cyan views
	StereoGeoWall		//	side-by-side views, to be superimposed by dual projectors
} StereoMode;


//	π
#define PI		3.14159265358979323846

//	square roots
#define ROOT2	1.41421356237309504880
#define ROOT3	1.73205080756887729352

//	The main tiling panel's size (width or height) in intrinsic units
//	does not depend on the perspective measurements that the user has reported,
//	but it does depend slightly on the geometry.
#define TILING_PANEL_HALF_SIZE_INTRINSIC_SPH	1.2
#define TILING_PANEL_HALF_SIZE_INTRINSIC_EUC	1.0
#define TILING_PANEL_HALF_SIZE_INTRINSIC_HYP	1.0

//	TILING_PANEL_RADIUS_EUC gives an upper bound on the distance 
//	in intrinsic units from the panel's center to one of its corners.
//	In the typical case that the panel is square, the bound is exact.
#define TILING_PANEL_RADIUS_EUC		( ROOT2 * TILING_PANEL_HALF_SIZE_INTRINSIC_EUC )

//	How many standard textures are available?
#define NUM_STANDARD_TEXTURES	10

//	The background image painting information is stored as if
//	it were a fourth face.  In other words, indices 0, 1 and 2
//	refer to real faces, while index 3 refers to the background.
#define BACKGROUND_IMAGE	3


//	Draw the tiling as a polyhedron with flat faces
//	or as a tiling of a curved surface?
typedef enum
{
	StyleFlat,
	StyleCurved
} ktTilingStyle;

//	KaleidoTile presents a Hue-Saturation-Lightness-Opacity (HSLA)
//	color model to the user, but converts to Red-Green-Blue-Opacity (RGBA)
//	when communicating with OpenGL.

typedef struct
{
	double	h,
			s,
			l,
			a;
} ColorHSLA;

typedef struct
{
	double	r,	//	red,   premultiplied by alpha
			g,	//	green, premultiplied by alpha
			b,	//	blue,  premultiplied by alpha
			a;	//	alpha = opacity
} ColorRGBA;

typedef struct
{
	Byte	r,	//	red,   premultiplied by alpha
			g,	//	green, premultiplied by alpha
			b,	//	blue,  premultiplied by alpha
			a;	//	alpha = opacity
} PixelRGBA;

//	The three face classes and the background may each be colored, textured or invisible.  
typedef enum
{
	//	MouseDown() and RenderControlPanel() rely on 
	//	the specific values of these enumeration constants.
	FacePaintColorHS	= 0,	//	FacePaintColorHS and FacePaintColorLA differ only in which
	FacePaintColorLA,			//		components of the HSLA color are current being edited.
	FacePaintTextureStandard,	//	one of the built-in textures
	FacePaintTextureCustom,		//	a user-provided texture
	FacePaintInvisible			//	face or background isn't drawn at all
} ktFacePaintStyle;
typedef struct
{
	//	Is the face or background invisible, colored or textured?
	ktFacePaintStyle	itsStyle;

	//	If the face is colored, what color?
	ColorHSLA			itsHSLA;	//	hue, saturation, lightness, opacity
	
	//	If the face (or background) has a standard texture, what is its ID?
	//	Standard texture IDs run from 1 through NUM_STANDARD_TEXTURES.
	//	Note:  The actual textures will be different for the background
	//	than for the faces of the tiling, even though the range of itsStandardTextureID
	//	is the same in both cases.
	unsigned int		itsStandardTextureID;
	
	//	If the face has a custom texture, what are its pixels?
	//	A custom texture is always square, with power of two dimensions.
	//	When no custom texture is present, keep itsCustomTexturePixels set to NULL.
	//	A custom texture may, however, be present even when it's not currently
	//	being displayed.  KaleidoTile keeps it around in case the user wants
	//	to return to it.
	unsigned int		itsCustomTextureSize;			//	must be a power of two
	double				itsCustomTextureFraction,		//	what fraction of the power-of-two square contains real pixels
														//		(used only for face textures, ignored for background textures)
						itsCustomTextureAspectRatio;	//	width / height
														//		(used only for background textures, ignored for face textures)
	PixelRGBA			*itsCustomTexturePixels;		//	linear (row major) array of RGBA pixels

} ktFacePainting;

//	Where might a mouse click fall?
#define NUM_SELECTION_BUTTONS	5
typedef enum
{
	MouseTargetNone,
	MouseTargetTiling,
	MouseTargetSymmetryBase,
	MouseTargetSymmetry3 = MouseTargetSymmetryBase,
	MouseTargetSymmetry4,
	MouseTargetSymmetry5,
	MouseTargetSymmetry6,
	MouseTargetSymmetry7,
	MouseTargetStyleFlat,
	MouseTargetStyleCurved,
	MouseTargetTriangle,
	MouseTargetSelectBase,
	MouseTargetSelectABase = (MouseTargetSelectBase + 0*NUM_SELECTION_BUTTONS),
	MouseTargetSelectAColorHS = MouseTargetSelectABase,
	MouseTargetSelectAColorLA,
	MouseTargetSelectATextureStandard,
	MouseTargetSelectATextureCustom,
	MouseTargetSelectAInvisible,
	MouseTargetSelectBBase = (MouseTargetSelectBase + 1*NUM_SELECTION_BUTTONS),
	MouseTargetSelectBColorHS = MouseTargetSelectBBase,
	MouseTargetSelectBColorLA,
	MouseTargetSelectBTextureStandard,
	MouseTargetSelectBTextureCustom,
	MouseTargetSelectBInvisible,
	MouseTargetSelectCBase = (MouseTargetSelectBase + 2*NUM_SELECTION_BUTTONS),
	MouseTargetSelectCColorHS = MouseTargetSelectCBase,
	MouseTargetSelectCColorLA,
	MouseTargetSelectCTextureStandard,
	MouseTargetSelectCTextureCustom,
	MouseTargetSelectCInvisible,
	MouseTargetSelectBackgroundBase = (MouseTargetSelectBase + 3*NUM_SELECTION_BUTTONS),
	MouseTargetSelectBackgroundColorHS = MouseTargetSelectBackgroundBase,
	MouseTargetSelectBackgroundColorLA,
	MouseTargetSelectBackgroundTextureStandard,
	MouseTargetSelectBackgroundTextureCustom,
	MouseTargetSelectBackgroundInvisible,
	MouseTargetPaintBase,
	MouseTargetPaintA = MouseTargetPaintBase,
	MouseTargetPaintB,
	MouseTargetPaintC,
	MouseTargetPaintBackground,
	MouseTargetGeneric
} MouseTarget;


//	All platform-independent data about the space and
//	how it's displayed live in the ModelData structure.
typedef struct
{
	//	Have the various OpenGL elements been pre-initialized?
	bool			itsPreparedExtensions,
					itsPreparedShaders,
					itsPreparedTextures;
	
	//	Which OpenGL extensions does the current context support?
	bool			itsMultisamplingAvailableFlag,
					itsAnisotropicAvailableFlag,
					itsShadersAvailableFlag;
	
	//	Which OpenGL extensions would the user like enabled?
	bool			itsMultisamplingEnabledFlag,
					itsAnisotropicEnabledFlag,
					itsShadersEnabledFlag;
				
	//	OpenGL shaders, textures and vertex/index buffers.
	unsigned int	itsShaderPrograms[NumShaders],
					itsTextureNames[NumTextures],
					itsVertexBufferNames[NumVertexBuffers],
					itsIndexBufferNames [NumVertexBuffers];

	//	Which images to show?
	bool			itsPositiveImagesFlag,
					itsNegativeImagesFlag;
	
	//	Cut along mirror lines?
	bool			itsMirrorLinesFlag;
	
	//	Snap to Archimedean solids?
	bool			itsSnapToArchimedeanFlag;
	const char		*itsArchimedeanKey;	//	zero-terminated UTF-8 string, NULL when no name is shown
	double			itsArchimedeanNameDisplayEndTime;
	
	//	Stereo 3D mode
	StereoMode		itsStereoMode;

	//	Synchronize to the monitor's refresh rate?
	bool			itsSyncFlag;

	//	Environmental measurements permit accurate simulation.
	//	These quantities correspond to the variables d, w, h, e and c
	//	in the file FlatSpaceProjectionMatrices.pdf.
	double			itsWindowDistance,
					itsWindowHalfWidth,
					itsWindowHalfHeight,
					itsEyeOffset;		//	for stereo 3D only
	
	//	When some part of the program (for example the mouse-handling code)
	//	wants to request a redraw, it sets itsRedrawRequestFlag.
	//	The idle-time routine will handle the request and clear the flag.
	//
	//	Note:  It's essential that all redraws funnel through the idle-time routine.
	//	For example, if mouse movements generated redraw events directly, this could
	//	saturate the event loop, thus blocking normal idle-time motion.
	bool			itsRedrawRequestFlag;
	
	//	Keep track of ModelView and Projection matrices.
	double			itsModelViewMatrix[4][4],
					itsProjectionMatrix[4][4];

	//	Purely geometrical description of the current tiling.
	ktTiling		itsTiling;
	
	//	Store an instantaneous velocity for the tiling.
	//	The instantaneous velocity may be scaled to fit any time interval.
	ktVelocity		itsTilingVelocity;
	
	//	Store the tiling's current position as a Lie derivative.
	//	The true position is obtained by integrating this "positional
	//	derivative" for unit time.  Typically the position will keep
	//	the north pole (0,0,1) fairly close to its original location,
	//	so we may retain the same "positional derivative" even when
	//	changing geometries.
	ktVelocity		itsTilingPosition;
	
	//	Store the triple point location in barycentric coordinates,
	//	defined so that (1,0,0), (0,1,0) and (0,0,1) are the fundamental
	//	triangle's vertices;  (0, 1/2, 1/2), (1/2, 0, 1/2) and (1/2, 1/2, 0)
	//	are the intersections of angle bisectors with opposite sides;
	//	and (1/3, 1/3, 1/3) is the north pole.  Note that the components
	//	of itsTriplePointPosition must always sum to 1.
	ktVector		itsTriplePointPosition;
	
	//	Store the triple point velocity relative to the same
	//	barycentric coordinates of itsTriplePointPosition.
	//	Note that the components of itsTriplePointVelocity
	//	must always sum to 0.
	ktVector		itsTriplePointVelocity;
	
	//	Flat faces or curved faces?
	ktTilingStyle	itsTilingStyle;
	
	//	Colored, textured or invisible faces and background?
	//	The indices 0, 1 and 2 refer to faces on the polyhedron,
	//	while the index 3 refers to the background image.
	//	For clarity, BACKGROUND_IMAGE is #defined to be 3.
	ktFacePainting	itsFacePainting[4];
	
	//	How fine a mesh should approximate each curved quad?
	unsigned int	itsMeshSize;
	
	//	How deep should the hyperbolic tiling go?
	double			itsHyperbolicTilingRadius;	//	in radians, typically ∈ [4.0, 6.0]

	//	Keep track of where a drag begins.
	MouseTarget		itsMouseTarget;
	double			itsPreviousMousePH,	//	for updating position
					itsPreviousMousePV,
					itsPreviousMouseVH,	//	for updating velocity
					itsPreviousMouseVV;
	double			itsPreviousMouseVTime;	//	timestamp in seconds
	ktVelocity		itsNewTilingVelocity;
	ktVector		itsNewTriplePointVelocity;

} ModelData;


//	Ordinary rendering uses a single viewpoint,
//	while anaglyphic 3D uses separate left- and right-eye views.
typedef enum
{
	EyeOnly,
	EyeLeft,
	EyeRight
} EyeType;


//	A dictionary will translate all user-visible phrases
//	to the currently active language.  
//	The dictionary's structure is private to KaleidoTileLocalization.c.
//	For all other purposes the following opaque typedef suffices.
typedef struct LocalizedPhrase	LocalizedPhrase;


//	Platform-independent global variables

//	Two-letter language code, such as "en" or "ja".
extern char				gLanguageCode[3];	//	zero-terminated ASCII string

//	Keep a dictionary for translating all user-interface text
//	into the current language.
extern LocalizedPhrase	*gDictionary;

//	Show frame rate?
extern bool				gShowFrameRate;

//	Test for memory leaks.
extern signed int		gMemCount;


//	Platform-dependent global functions,
//	in KaleidoTile-Win32-Utilities.c or KaleidoTile-MacOS-Utilities.c
extern bool			GetBasePath(char *aPath, unsigned int aPathSize);
extern ErrorText	GetFileContents(const char *aPathName, unsigned int *aNumRawBytes, Byte **someRawBytes);
extern void			FreeFileContents(unsigned int *aNumRawBytes, Byte **someRawBytes);
extern void			SetAlphaTextureFromString(unsigned int aTextureName, const char *aString, 
						unsigned int aWidth, unsigned int aHeight, 
						char *aFontName, unsigned int aFontSize, unsigned int aFontDescent,
						bool aCenteringFlag, unsigned int aMargin,
						bool aShadersFlag, ErrorText *aFirstError);
extern bool			GetImageToPaste(unsigned int *aWidth, unsigned int *aHeight, PixelRGBA **aPixelArray);
extern void			FreeImageToPaste(PixelRGBA **aPixelArray);
extern void			PlayWAV(char *aRelativePath);
extern bool			GetUserPrefBool(char *aKey);
extern void			SetUserPrefBool(char *aKey, bool aValue);
extern void			ErrorMessage(ErrorText aMessage, ErrorText aTitle);

//	Platform-independent global functions

//	in KaleidoTileInit.c
extern bool			ShadersAreAvailable(void);
extern void			SetUpModelData(ModelData *md);
extern void			ShutDownModelData(ModelData *md);
extern ErrorText	SetUpGraphicsAsNeeded(ModelData *md);
extern void			ShutDownGraphicsAsNeeded(ModelData *md);

//	in KaleidoTileUtilities.c
extern ErrorText	GetAbsolutePath(const char *aDirectory, const char *aFileName, char *aPath, unsigned int aPathSize);
extern void			ConvertColorToGreyscale(ColorRGBA *anRGBAColor);
extern void			ConvertPixelsToGreyscale(PixelRGBA *aPixelArray, unsigned int aLength);

//	in KaleidoTileOptions.c
extern void			SetPositiveImagesFlag(ModelData *md, bool aPositiveImagesFlag);
extern void			SetNegativeImagesFlag(ModelData *md, bool aNegativeImagesFlag);
extern void			SetMirrorLinesFlag(ModelData *md, bool aMirrorLinesFlag);
extern void			SetBiggerTiles(ModelData *md);
extern void			SetSmallerTiles(ModelData *md);
extern void			SetMultisamplingEnabledFlag(ModelData *md, bool aMultisamplingEnabledFlag);
extern void			SetStereo3DMode(ModelData *md, StereoMode aStereo3DMode);
extern void			SetMeshSize(ModelData *md, unsigned int aMeshSize);
extern void			SetHyperbolicTilingRadius(ModelData *md, double aTilingRadius);
extern void			SetAnisotropicEnabledFlag(ModelData *md, bool anAnisotropicEnabledFlag);
extern void			SetShadersEnabledFlag(ModelData *md, bool aShadersEnabledFlag);

//	in KaleidoTileSimulation.c
extern bool			SimulationWantsUpdates(ModelData *md);
extern void			SimulationUpdate(ModelData *md, double aFramePeriod, double aTimeStamp);
extern void			ChangeTiling(ModelData *md, unsigned int p, unsigned int q, unsigned int r);
extern void			ChangeStyle(ModelData *md, ktTilingStyle aStyle);
extern void			SetFaceColorHS(ktFacePainting *aFacePainting, double aHue, double aSaturation, double aLightness, double anOpacity);
extern void			SetFaceColorLA(ktFacePainting *aFacePainting, double aHue, double aSaturation, double aLightness, double anOpacity);
extern void			SetFaceTextureStandard(ktFacePainting *aFacePainting, unsigned int aStandardTextureID, bool *aPreparedTexturesFlag);
extern void			SetFaceTextureCustom(ktFacePainting *aFacePainting, bool aBackgroundImageFlag, bool aPanelClickFlag, bool *aPreparedTexturesFlag);
extern void			SetFaceInvisible(ktFacePainting *aFacePainting);

//	in KaleidoTileMouse.c
extern void			MouseDown(ModelData *md, double aMouseH, double aMouseV, double aTimeStamp);
extern void			MouseMove(ModelData *md, double aMouseH, double aMouseV, double aTimeStamp);
extern void			MouseUp  (ModelData *md, double aMouseH, double aMouseV, double aTimeStamp);

//	in KaleidoTileGraphics.c
extern char			*GetRendererName(void);
extern void			SetTextureImageFancy(unsigned int aTextureName, unsigned int aWidth, unsigned int aHeight, unsigned int aDepth, Byte *aPixelArray, bool aShadersFlag);
extern bool			IsPowerOfTwo(unsigned int n);
extern void			ResizeGLView(ModelData *md, unsigned int aNewViewWidth, unsigned int aNewViewHeight);
extern ErrorText	ReadRGBAPixels(unsigned int aWidth, unsigned int aHeight, Byte *aBuffer);
extern ErrorText	Render(ModelData *md);
extern MouseTarget	InterpretMousePosition(double *aMouseH, double *aMouseV, double aDomainScale, double aDomainOffset);
extern ErrorText	GetErrorString(void);

//	in KaleidoTileLocalization.c
extern void			RefreshLocalization(const char aTwoLetterLanguageCode[3]);
extern const char	*GetLocalizedText(const char *aKey);
