//	KaleidoTileInit.c
//
//	Initializes the ModelData.
//	Prepares OpenGL textures and display lists.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Common.h"
#include <math.h>
#include <time.h>	//	for seeding srand()

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif

//	How fast should the tiling be moving initially?
#define INITIAL_TILING_SPEED	0.03

//	How fast should the triple point be moving initially?
#define INITIAL_TRIPLE_POINT_SPEED	0.02


//	RGBA image, as read from a .rgba file
typedef struct
{
	//	Keep a pointer to the raw data, including the header bytes.
	unsigned int	itsNumRawBytes;
	Byte			*itsRawBytes;
	
	//	Interpret the data as a width, a height and some pixels.
	unsigned int	itsWidth,
					itsHeight;
	PixelRGBA		*itsPixels;	//	= itsRawBytes + 8

} ImageRGBA;

//	Texture coordinates may wrap or clamp.
typedef enum
{
	TextureWrap,
	TextureClamp
} TextureBoundary;

//	Textures may reduce to greyscale for anaglyphic 3D.
typedef enum
{
	AnaglyphColor,
	AnaglyphGrey
} AnaglyphRange;


static ErrorText		SetUpExtensions(ModelData *md);
static bool				WholeWordSubstring(const char *aString, const char *aPotentialSubstring);
static void				ShutDownExtensions(ModelData *md);
static ErrorText		SetUpShaders(ModelData *md);
static ErrorText		SetUpOneShaderProgram(unsigned int *aShaderProgram, ShaderIndex aShaderIndex);
static void				ShutDownShaders(ModelData *md);
static ErrorText		SetUpTextures(ModelData *md);
static void				SetUpOneTexture(ModelData *md, unsigned int *aTextureName, char *aTextureFileName, TextureBoundary aTextureBoundary, AnaglyphRange anAnaglyphRange, ErrorText *aFirstError);
static void				LoadPaintingTexture(ModelData *md, unsigned int aTextureName, unsigned int aFaceIndex, AnaglyphRange anAnaglyphRange, ErrorText *aFirstError);
static void				SetCheckerboardTexels(unsigned int aTextureName);
static ErrorText		ReadImageRGBA(char *aTextureFileName, bool aGreyscaleFlag, ImageRGBA **anImageRGBA);
static void				FreeImageRGBA(ImageRGBA **anImage);
static void				ShutDownTextures(ModelData *md);


bool ShadersAreAvailable(void)
{
	const char	*theExtensionsString;

	//	Assume the OpenGL context has already been set.
	
	theExtensionsString = (const char *) glGetString(GL_EXTENSIONS);

	if (theExtensionsString != NULL)
		return
			WholeWordSubstring(theExtensionsString, "GL_ARB_shader_objects" )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_vertex_shader"  )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_fragment_shader");
	else
		return false;
}


void SetUpModelData(ModelData *md)
{
	double			r,
					z,
					theta;
	unsigned int	i;

	//	All OpenGL elements require initialization.
	md->itsPreparedExtensions	= false;
	md->itsPreparedShaders		= false;
	md->itsPreparedTextures		= false;

	//	We'll test for OpenGL extensions at the last minute,
	//	just before rendering the first frame.
	//	For now set safe defaults.
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;
	
	//	No shaders, textures or vertex/index buffers are present yet.

	for (i = 0; i < NumShaders; i++)
		md->itsShaderPrograms[i] = 0;

	for (i = 0; i < NumTextures; i++)
		md->itsTextureNames[i] = 0;

	for (i = 0; i < NumVertexBuffers; i++)
	{
		md->itsVertexBufferNames[i] = 0;
		md->itsIndexBufferNames [i] = 0;
	}

	//	Leave stereo off by default.
	md->itsStereoMode	= StereoNone;

	//	Enable synchronization by default.
	md->itsSyncFlag		= true;

	//	Multisampling takes place iff itsMultisamplingAvailableFlag
	//	and itsMultisamplingEnabledFlag are both true.
	//
	//	KaleidoTileMacOSView's initWithFrame method (Mac)
	//	or GetUserPrefBool() itself (Win32) sets the default value.
	md->itsMultisamplingEnabledFlag	= GetUserPrefBool("multisampling");

	//	Enable anisotropic texture filtering by default.
	//	(Anisotropic texture filtering takes place iff
	//	itsAnisotropicAvailableFlag and itsAnisotropicEnabledFlag
	//	are both true.)
	md->itsAnisotropicEnabledFlag	= true;
	
	//	Shaders are used iff itsShadersAvailableFlag 
	//	and itsShadersEnabledFlag are both true.
	//
	//	KaleidoTileMacOSView's initWithFrame method (Mac)
	//	or GetUserPrefBool() itself (Win32) sets the default value.
//	md->itsShadersEnabledFlag		= GetUserPref("shaders");
//IGNORE USER PREF AND ALWAYS START WITH SHADERS DISABLED,
//FOR MAXIMAL ROBUSTNESS IN THIS ARCHIVED VERSION OF THE PROGRAM.
md->itsShadersEnabledFlag = false;

	//	Set up window dimensions.
	//	These dimensions refer, of course, to the square main view,
	//	excluding the control panel.
	//
	//	For a realistic simulation, the program's internal parameters
	//	must match the relative proportions of the user's physical display device,
	//	the distance from the display to the user's face, and the separation
	//	between the user's eyes.
	//	At runtime the user may use the Preferences panel
	//	to override the defaults given here.
	md->itsWindowDistance	= 40.0;	//	in centimeters
	md->itsWindowHalfWidth	= 10.0;	//	in centimeters
	md->itsWindowHalfHeight	= 10.0;	//	in centimeters
	md->itsEyeOffset		=  2.5;	//	nose to eye

	//	No redraw request is pending.
	md->itsRedrawRequestFlag	= false;


	//	Set up application-specific data.

	//	Show images of both chiralities.
	md->itsPositiveImagesFlag	= true;
	md->itsNegativeImagesFlag	= true;

	//	Don't cut along mirror lines.
	md->itsMirrorLinesFlag	= false;

	//	Enable Snap to Archimedean Solids by default.
	//	Show but don't say their names.
	md->itsSnapToArchimedeanFlag			= true;
	md->itsArchimedeanKey					= NULL;
	md->itsArchimedeanNameDisplayEndTime	= 0.0;

	//	Choose a mesh size.  4 should be fine for most purposes.
	md->itsMeshSize	= 4;
	
	//	Choose a hyperbolic tiling radius.
	//	For most purposes, radis 4.5 looks good and runs fast.
	md->itsHyperbolicTilingRadius = 4.5;

	//	Set up a tiling.
	SetUpTiling(&md->itsTiling, 2, 3, 5, 0);

	//	Set up an initial tiling position.
	md->itsTilingPosition.x = 0.0;
	md->itsTilingPosition.y = 0.0;
	md->itsTilingPosition.r = 0.0;
	
	//	Initialize the random number generator rand().
	//	rand() produces numbers that are not very random;
	//	for details, see comments preceding srand(time(NULL)) in Torus Games.
	srand(time(NULL));

	//	Set up an initial tiling velocity.
	//	To get a uniform distribution on the unit sphere in velocity space,
	//	choose coordinates (theta, z) on a circumscribed cylinder
	//	and project cylindrically onto the unit sphere.
	z		= 2.0 * (double)rand() / (double)RAND_MAX  -  1.0;
	r		= sqrt(1.0 - z*z);
	theta	= 2.0 * PI * (double)rand() / (double)RAND_MAX;
	md->itsTilingVelocity.x = INITIAL_TILING_SPEED * r * cos(theta);
	md->itsTilingVelocity.y = INITIAL_TILING_SPEED * r * sin(theta);
	md->itsTilingVelocity.r = INITIAL_TILING_SPEED * z;

	//	Set up an initial triple point position.
	md->itsTriplePointPosition[0] = 1.0/3.0;
	md->itsTriplePointPosition[1] = 1.0/3.0;
	md->itsTriplePointPosition[2] = 1.0/3.0;

	//	Set up an initial triple point velocity.
	theta	= 2.0 * PI * (double)rand() / (double)RAND_MAX;
	md->itsTriplePointVelocity[0] = INITIAL_TRIPLE_POINT_SPEED * (cos(theta)*(-sqrt(1.0/6.0)) + sin(theta)*(+sqrt(1.0/2.0)));
	md->itsTriplePointVelocity[1] = INITIAL_TRIPLE_POINT_SPEED * (cos(theta)*(-sqrt(1.0/6.0)) + sin(theta)*(-sqrt(1.0/2.0)));
	md->itsTriplePointVelocity[2] = INITIAL_TRIPLE_POINT_SPEED * (cos(theta)*(+sqrt(4.0/6.0)) + sin(theta)*(     0.0      ));

	//	Choose a style.
	md->itsTilingStyle = StyleFlat;

	//	Initialize the paintings to "empty" values.
	//	Real colors and/or textures will get set momentarily.
	for (i = 0; i < 4; i++)
	{
		//	Set a reasonable color default, so that, for example,
		//	if we initialize the painting to a texture (see below),
		//	and the user subsequently switches to the hue-saturation palette,
		//	s/he will see something reasonable.
		md->itsFacePainting[i].itsHSLA.h = 0.0;
		md->itsFacePainting[i].itsHSLA.s = 1.0;
		md->itsFacePainting[i].itsHSLA.l = 0.5;
		md->itsFacePainting[i].itsHSLA.a = 1.0;
		
		//	Similarly, initialize itsStandardTextureID to some reasonable value.
		md->itsFacePainting[i].itsStandardTextureID = 1;

		//	Initialize custom texture pointer to NULL,
		//	so SetFaceColor*() and/or SetFaceTextureStandard()
		//	(see below) will know that no pixel buffer is present.
		md->itsFacePainting[i].itsCustomTextureSize			= 0;
		md->itsFacePainting[i].itsCustomTextureFraction		= 1.0;
		md->itsFacePainting[i].itsCustomTextureAspectRatio	= 1.0;
		md->itsFacePainting[i].itsCustomTexturePixels		= NULL;
	}

	//	Choose face and background colors or textures.
#ifdef FOR_TALKS
	SetFaceColorHS(&md->itsFacePainting[0], 0.00, 0.75, 0.50, 1.00);
	SetFaceColorLA(&md->itsFacePainting[1], 0.60, 1.00, 0.67, 1.00);
	md->itsFacePainting[2].itsHSLA.h = 0.67;
	SetFaceTextureStandard(&md->itsFacePainting[2], 4, &md->itsPreparedTextures);
	SetFaceTextureStandard(&md->itsFacePainting[BACKGROUND_IMAGE], 2, &md->itsPreparedTextures);
#else
	SetFaceColorHS(&md->itsFacePainting[0], 0.00, 0.75, 0.50, 1.00);
	SetFaceColorLA(&md->itsFacePainting[1], 0.60, 1.00, 0.67, 0.60);
	md->itsFacePainting[2].itsHSLA.h = 0.67;
	SetFaceTextureStandard(&md->itsFacePainting[2], 9, &md->itsPreparedTextures);
	SetFaceTextureStandard(&md->itsFacePainting[BACKGROUND_IMAGE], 3, &md->itsPreparedTextures);
#endif
}


void ShutDownModelData(ModelData *md)
{
	unsigned int	i;

	for (i = 0; i < 4; i++)
		FREE_MEMORY_SAFELY(md->itsFacePainting[i].itsCustomTexturePixels);
}


ErrorText SetUpGraphicsAsNeeded(ModelData *md)
{
	//	These are one-time initializations,
	//	so usually no work is required here.
	//	Only in exceptional circumstances, for example
	//	when the user has changed the contents or display settings,
	//	will something need an update.

	//	Assume the OpenGL context has already been set.
	
	ErrorText	theError;

	if ( ! md->itsPreparedExtensions )
	{
		if ((theError = SetUpExtensions(md)) != NULL)
			return theError;
		md->itsPreparedExtensions = true;

		//	Availability of extensions may influence
		//	the construction of textures, etc.,
		//	so rebuild everything from scratch.
		md->itsPreparedShaders	= false;
		md->itsPreparedTextures	= false;
	}

	if ( ! md->itsPreparedShaders )
	{
		if ((theError = SetUpShaders(md)) != NULL)
			return theError;
		md->itsPreparedShaders = true;
		
		//	Alpha-only textures get handled differently
		//	in the OpenGL 1.1 fixed-function pipeline,
		//	in OpenGL 2.0 shaders, and in OpenGL 3.0 shaders.
		//	Please see the comments in SetTextureImageFancy()
		//	for details.
		md->itsPreparedTextures = false;
	}

	if ( ! md->itsPreparedTextures )
	{
		if ((theError = SetUpTextures(md)) != NULL)
			return theError;
		md->itsPreparedTextures	= true;
	}
	
	return NULL;
}


void ShutDownGraphicsAsNeeded(ModelData *md)
{
	//	Assume the OpenGL context has already been set.

	//	Shut down the various elements in the opposite order
	//	from that in which they were created.  In particular,
	//	be sure to shut down the extensions last, so the other
	//	shutdown routines will know what extensions have been
	//	available.  For example, ShutDownShaders() will need
	//	to know whether shaders have been available or not.
	ShutDownTextures(md);
	ShutDownShaders(md);
	ShutDownExtensions(md);

	md->itsPreparedExtensions	= false;
	md->itsPreparedShaders		= false;
	md->itsPreparedTextures		= false;
}


static ErrorText SetUpExtensions(ModelData *md)
{
	const char	*theExtensionsString;

	//	Desired OpenGL extensions are assumed absent until found.
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;

	//	Get the renderer's extensions string.
	theExtensionsString = (const char *) glGetString(GL_EXTENSIONS);

	if (theExtensionsString != NULL)
	{
		//	Is multisampling available?
		if (WholeWordSubstring(theExtensionsString, "GL_ARB_multisample"))
			md->itsMultisamplingAvailableFlag = true;

		//	Is anisotropic texture filtering available?
		if (WholeWordSubstring(theExtensionsString, "GL_EXT_texture_filter_anisotropic"))
			md->itsAnisotropicAvailableFlag = true;

		//	Are shaders available?
		if (WholeWordSubstring(theExtensionsString, "GL_ARB_shader_objects" )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_vertex_shader"  )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_fragment_shader"))
			md->itsShadersAvailableFlag = true;
	}
	
	//	Did any OpenGL errors occur?
	return GetErrorString();
}


static bool WholeWordSubstring(
	const char	*aString,				//	null-terminated
	const char	*aPotentialSubstring)	//	null-terminated
{
	const char	*thePotentialMatch,
				*s,
				*p;
	bool		theMatchFlag;

	//	Does aPotentialSubstring occur as a complete space-delimited word in aString?

	//	Irrelevant technical detail:  an empty aPotentialSubstring
	//	(consisting of a terminating zero alone) counts as a substring
	//	of any string, even an empty one.

	//	Consider the first character in each space-delimited word of aString
	//	as the starting point for a potential match.
	thePotentialMatch = aString;
	while (*thePotentialMatch == ' ')
		thePotentialMatch++;
	do
	{
		//	Compare each character in aPotentialSubstring
		//	(excluding the terminating zero) to thePotentialMatch.
		s = thePotentialMatch;
		p = aPotentialSubstring;
		theMatchFlag = true;
		while (theMatchFlag && *p != 0)
		{
			if (*p++ != *s++)
				theMatchFlag = false;
		}

		//	If a space or terminating 0 follows a match, return true.
		if (theMatchFlag
		 && (*s == ' ' || *s == 0))
			return true;

		//	Skip to the beginning of the next whole word in thePotentialMatch.
		//		Skip by non-spaces.
		while (*thePotentialMatch != ' ' && *thePotentialMatch != 0)
			thePotentialMatch++;
		//		Skip by intervening spaces.
		while (*thePotentialMatch == ' ')
			thePotentialMatch++;

		//	Continue with the loop to examine the next whole word (if any).

	} while (*thePotentialMatch != 0);

	return false;
}


static void ShutDownExtensions(ModelData *md)
{
	//	Just for good form...
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;
}


static ErrorText SetUpShaders(ModelData *md)
{
	ErrorText		theError;
	unsigned int	i;

	if (md->itsShadersAvailableFlag)
	{
		//	Delete any pre-existing shaders.
		ShutDownShaders(md);

		//	If shaders are desired, then...
		if (md->itsShadersEnabledFlag)
		{
			//	...create a shader of each type.
			//
			//	At run time, swapping simple shaders in and out as needed turns out
			//	to be faster than running a single all-purpose shader with options,
			//	because (much to my surprise) an if/then block containing
			//	texture sampling slows the fragment shader to a crawl
			//	(at least on my Radeon X1600).
			for (i = 0; i < NumShaders; i++)
			{
				theError = SetUpOneShaderProgram(&md->itsShaderPrograms[i], (ShaderIndex) i );
				if (theError != NULL)
					return theError;
			}
		}
	}
	
	//	Did any OpenGL errors occur?
	return GetErrorString();
}


static ErrorText SetUpOneShaderProgram(
	unsigned int	*aShaderProgram,	//	output
	ShaderIndex		aShaderIndex)		//	input
{
	ErrorText		theErrorMessage					= NULL;
	char			*theVertexShaderFileName		= NULL,
					*theFragmentShaderFileName		= NULL,
					theFilePath[2048];
	unsigned int	theVertexShaderSourceLength		= 0,
					theFragmentShaderSourceLength	= 0;
	Byte			*theVertexShaderSource			= NULL,
					*theFragmentShaderSource		= NULL;
	const char		*theVertexShaderPrefix			= NULL,
					*theFragmentShaderPrefix		= NULL;
	unsigned int	theVertexShader					= 0,
					theFragmentShader				= 0;
	GLint			theSuccessFlag;

	static GLchar	theInfoLog[1024] = {0};	//	static so it persists after this function returns
	
	//	For safe error handling.
	//	(*aShaderProgram should already be 0, but reset it anyway.)
	*aShaderProgram = 0;
	
	//	Select an appropriate pair of shaders for the given ShaderIndex.
	switch (aShaderIndex)
	{
		case ShaderColor:
			theVertexShaderFileName		= "KaleidoTile-Color.vs";
			theFragmentShaderFileName	= "KaleidoTile-Color.fs";
			break;

		case ShaderTexture:
			theVertexShaderFileName		= "KaleidoTile-Texture.vs";
			theFragmentShaderFileName	= "KaleidoTile-Texture.fs";
			break;

		case ShaderColorWithTexture:
			theVertexShaderFileName		= "KaleidoTile-ColorWithTexture.vs";
			theFragmentShaderFileName	= "KaleidoTile-ColorWithTexture.fs";
			break;

		case ShaderColorWithAlphaMask:
			theVertexShaderFileName		= "KaleidoTile-ColorWithAlphaMask.vs";
			theFragmentShaderFileName	= "KaleidoTile-ColorWithAlphaMask.fs";
			break;

		case ShaderColorWithLighting:
			theVertexShaderFileName		= "KaleidoTile-ColorWithLighting.vs";
			theFragmentShaderFileName	= "KaleidoTile-ColorWithLighting.fs";
			break;

		case ShaderTextureWithLighting:
			theVertexShaderFileName		= "KaleidoTile-TextureWithLighting.vs";
			theFragmentShaderFileName	= "KaleidoTile-TextureWithLighting.fs";
			break;

		case ShaderHSLAWithTexture:
			theVertexShaderFileName		= "KaleidoTile-HSLAWithTexture.vs";
			theFragmentShaderFileName	= "KaleidoTile-HSLAWithTexture.fs";
			break;

		default:
			theErrorMessage = "Invalid shader index";
			goto CleanUpSetUpOneShaderProgram;
	}

	//	Load the vertex shader source code.
	if ((theErrorMessage = GetAbsolutePath(	"Shaders", 
											theVertexShaderFileName,
											theFilePath,
											BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	if ((theErrorMessage = GetFileContents(theFilePath, &theVertexShaderSourceLength, &theVertexShaderSource)) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	
	//	Load the fragment shader source code.
	if ((theErrorMessage = GetAbsolutePath(	"Shaders",
											theFragmentShaderFileName,
											theFilePath,
											BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	if ((theErrorMessage = GetFileContents(theFilePath, &theFragmentShaderSourceLength, &theFragmentShaderSource)) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	
	//	Select appropriate prefixes.
	if (RUNNING_OPENGL_3)
	{
		theVertexShaderPrefix	= "#version 130\n";
		theFragmentShaderPrefix	= "#version 130\n";
	}
	else
	{
		//	Convert GLSL 1.3 storage qualifiers "in" and "out"
		//	to their GLSL 1.1 equivalents "attribute" and "varying".
		theVertexShaderPrefix	= "#version 110\n#define in attribute\n#define out varying\n";
		theFragmentShaderPrefix	= "#version 110\n#define in varying\n";
	}

	//	Create the vertex shader.
	theVertexShader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(	theVertexShader,
					2,
					(const char *[2])	{	theVertexShaderPrefix,	(const char *)theVertexShaderSource	},
					(GLint [2])			{	-1,						theVertexShaderSourceLength			});
	glCompileShader(theVertexShader);
	glGetShaderiv(theVertexShader, GL_COMPILE_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetShaderInfoLog(theVertexShader, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

	//	Create the fragment shader.
	theFragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(	theFragmentShader,
					2,
					(const char *[2])	{	theFragmentShaderPrefix,	(const char *)theFragmentShaderSource	},
					(GLint [2])			{	-1,							theFragmentShaderSourceLength			});
	glCompileShader(theFragmentShader);
	glGetShaderiv(theFragmentShader, GL_COMPILE_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetShaderInfoLog(theFragmentShader, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

	//	Create the program and attach the vertex and fragment shaders.
	*aShaderProgram = glCreateProgram();
	if (*aShaderProgram == 0)
	{
		theErrorMessage = "Couldn't create *aShaderProgram";
		goto CleanUpSetUpOneShaderProgram;
	}
	glAttachShader(*aShaderProgram, theVertexShader  );
	glAttachShader(*aShaderProgram, theFragmentShader);
	
	//	Assign locations to vertex shader attributes.
	//	It's always safe to bind an attribute name, whether or not the given shader uses it.
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_POSITION,			"atrPosition"					);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_MVP_MATRIX_ROW_0,	"atrModelViewProjectionMatrix"	);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_DIFFUSE_EVALUATOR,	"atrDiffuseEvaluator"			);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_SPECULAR_EVALUATOR,	"atrSpecularEvaluator"			);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_NORMAL,				"atrNormal"						);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_TEX_COORD,			"atrTextureCoordinates"			);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_COLOR_HSLA,			"atrColorHSLA"					);
	
	//	Link the program.
	glLinkProgram(*aShaderProgram);
	glGetProgramiv(*aShaderProgram, GL_LINK_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetProgramInfoLog(*aShaderProgram, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

CleanUpSetUpOneShaderProgram:

	//	Free our private copy of the source code.
	FreeFileContents(&theVertexShaderSourceLength,   &theVertexShaderSource  );
	FreeFileContents(&theFragmentShaderSourceLength, &theFragmentShaderSource);

	//	Release our temporary references (if any) to the vertex and fragments shaders.
	//	The shader program will keep its own references to the vertex and fragments shaders,
	//	and will automatically delete them when it itself gets deleted.  
	//	Passing 0 is harmless.
	glDeleteShader(theVertexShader);
	glDeleteShader(theFragmentShader);

	//	If an error occured, delete the shader program (if any).
	if (theErrorMessage != NULL)
	{
		glDeleteProgram(*aShaderProgram);	//	OK to pass 0
		*aShaderProgram = 0;
	}

	return theErrorMessage;
}


static void ShutDownShaders(ModelData *md)
{
	unsigned int	i;

	if (md->itsShadersAvailableFlag)
	{
		glUseProgram(0);

		for (i = 0; i < NumShaders; i++)
		{
			glDeleteProgram(md->itsShaderPrograms[i]);
			md->itsShaderPrograms[i] = 0;
		}
	}
}


static ErrorText SetUpTextures(ModelData *md)
{
	ErrorText	theFirstError	= NULL;

	SetUpOneTexture(md, &md->itsTextureNames[TextureFaceA],				NULL,	TextureWrap,	AnaglyphGrey,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureFaceB],				NULL,	TextureWrap,	AnaglyphGrey,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureFaceC],				NULL,	TextureWrap,	AnaglyphGrey,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureTilingBackground],	NULL,	TextureWrap,	AnaglyphColor,	&theFirstError);
	LoadPaintingTexture(md, md->itsTextureNames[TextureFaceA],				0,					AnaglyphGrey,	&theFirstError);
	LoadPaintingTexture(md, md->itsTextureNames[TextureFaceB],				1,					AnaglyphGrey,	&theFirstError);
	LoadPaintingTexture(md, md->itsTextureNames[TextureFaceC],				2,					AnaglyphGrey,	&theFirstError);
	LoadPaintingTexture(md, md->itsTextureNames[TextureTilingBackground],	BACKGROUND_IMAGE,	AnaglyphColor,	&theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureRegularName],		NULL,	TextureClamp,	AnaglyphColor,	&theFirstError);
	if (md->itsArchimedeanKey != NULL)
	{
		SetAlphaTextureFromString(
			md->itsTextureNames[TextureRegularName],
			GetLocalizedText(md->itsArchimedeanKey),
			1024, 64, "Helvetica", 51, 13, true, 0,
			md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
			&theFirstError);
	}

	SetUpOneTexture(md, &md->itsTextureNames[TextureControlBackground],		"ControlPanelBackground.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
		
	//	If future translations (to other human languages) demand more horizontal space
	//	for the control panel labels, decrease aFontSize from its current value (21)
	//	to something smaller (perhaps 20).

	SetUpOneTexture(md, &md->itsTextureNames[TextureLabelSymmetry],			NULL,									TextureClamp,	AnaglyphColor,	&theFirstError);
	SetAlphaTextureFromString(md->itsTextureNames[TextureLabelSymmetry],
		GetLocalizedText("LabelSymmetry"),		256, 32, "Comic Sans MS", 21,  5, false, 16,
		md->itsShadersAvailableFlag && md->itsShadersEnabledFlag, &theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureLabelStyle],			NULL,									TextureClamp,	AnaglyphColor,	&theFirstError);
	SetAlphaTextureFromString(md->itsTextureNames[TextureLabelStyle],
		GetLocalizedText("LabelStyle"),			256, 32, "Comic Sans MS", 21,  5, false, 16,
		md->itsShadersAvailableFlag && md->itsShadersEnabledFlag, &theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureLabelControlPoint],		NULL,									TextureClamp,	AnaglyphColor,	&theFirstError);
	SetAlphaTextureFromString(md->itsTextureNames[TextureLabelControlPoint],
		GetLocalizedText("LabelControlPoint"),	256, 32, "Comic Sans MS", 21,  5, false, 16,
		md->itsShadersAvailableFlag && md->itsShadersEnabledFlag, &theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureLabelFaces],			NULL,									TextureClamp,	AnaglyphColor,	&theFirstError);
	SetAlphaTextureFromString(md->itsTextureNames[TextureLabelFaces],
		GetLocalizedText("LabelFaces"),			256, 32, "Comic Sans MS", 21,  5, false, 16,
		md->itsShadersAvailableFlag && md->itsShadersEnabledFlag, &theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureLabelBackground],		NULL,									TextureClamp,	AnaglyphColor,	&theFirstError);
	SetAlphaTextureFromString(md->itsTextureNames[TextureLabelBackground],
		GetLocalizedText("LabelBackground"),	256, 32, "Comic Sans MS", 21,  5, false, 16,
		md->itsShadersAvailableFlag && md->itsShadersEnabledFlag, &theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureSymmetry3],				"ControlPanelSymmetry3.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureSymmetry4],				"ControlPanelSymmetry4.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureSymmetry5],				"ControlPanelSymmetry5.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureSymmetry6],				"ControlPanelSymmetry6.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureSymmetry7],				"ControlPanelSymmetry7.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureStyleFlat],				"ControlPanelStyleFlat.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureStyleCurved],			"ControlPanelStyleCurved.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureRoundMarker],			"ControlPanelRoundMarker.rgba",			TextureWrap,	AnaglyphColor,	&theFirstError);

	SetUpOneTexture(md, &md->itsTextureNames[TextureBevel],					"ControlPanelBevel.rgba",				TextureClamp,	AnaglyphColor,	&theFirstError);

	//	The checkerboard texture is a special case, with no interpolation and no mipmapping.
	SetUpOneTexture(md, &md->itsTextureNames[TextureCheckerboard],			NULL,									TextureWrap,	AnaglyphColor,	&theFirstError);
	SetCheckerboardTexels(md->itsTextureNames[TextureCheckerboard]);

	SetUpOneTexture(md, &md->itsTextureNames[TextureSquareMarker],			"ControlPanelSquareMarker.rgba",		TextureWrap,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureFacePalette],			"ControlPanelFacePalette.rgba",			TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureBackgroundPalette],		"ControlPanelBackgroundPalette.rgba",	TextureClamp,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureButtonHueSaturation],	"ControlPanelButtonHS.rgba",			TextureWrap,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureButtonLightnessAlpha],	"ControlPanelButtonLA.rgba",			TextureWrap,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureButtonTexture],			"ControlPanelButtonTexStandard.rgba",	TextureWrap,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureButtonPaste],			"ControlPanelButtonTexCustom.rgba",		TextureWrap,	AnaglyphColor,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNames[TextureButtonNull],			"ControlPanelButtonNone.rgba",			TextureWrap,	AnaglyphColor,	&theFirstError);

	return theFirstError;

	//	Did any OpenGL errors occur?
	return GetErrorString();
}


static void SetUpOneTexture(
	ModelData		*md,
	unsigned int	*aTextureName,		//	output
	char			*aTextureFileName,	//	pass NULL to set up an empty texture
	TextureBoundary	aTextureBoundary,
	AnaglyphRange	anAnaglyphRange,
	ErrorText		*aFirstError)
{
	ErrorText	theErrorMessage		= NULL;
	float		theMaxAnisotropy	= 1.0;
	ImageRGBA	*theImageRGBA		= NULL;
	
	//	Release any pre-existing texture.
	glDeleteTextures(1, (GLuint *) aTextureName);	//	silently ignores *aTextureName == 0
	*aTextureName = 0;
	
	//	Generate a new texture name.
	glGenTextures(1, (GLuint *) aTextureName);

	//	Bind the texture.
	glBindTexture(GL_TEXTURE_2D, *aTextureName);

	//	Set minification and magnification filters.
	//	Mipmapping is almost always a good idea.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR               );

	//	Set anisotropic texture filtering.
	if (md->itsAnisotropicAvailableFlag)
	{
		if (md->itsAnisotropicEnabledFlag)
			glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &theMaxAnisotropy);	//	Allow the maximum.
		else
			theMaxAnisotropy = 1.0f;	//	Suppress anisotropic texture filtering.
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, theMaxAnisotropy);
	}

	//	Set wrapping or clamping.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, aTextureBoundary == TextureClamp ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, aTextureBoundary == TextureClamp ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	
	//	If the caller didn't want to load a file, we're done.
	if (aTextureFileName == NULL)
		return;

	//	Read the texture file.
	if ((theErrorMessage = ReadImageRGBA(	aTextureFileName,
											md->itsStereoMode == StereoAnaglyphic
											 && anAnaglyphRange == AnaglyphGrey,
											&theImageRGBA)) != NULL)
		goto CleanUpSetUpOneTexture;

	//	Pass the pixel data to OpenGL.
	SetTextureImageFancy(	*aTextureName,
							theImageRGBA->itsWidth,
							theImageRGBA->itsHeight,
							4,
							(Byte *)theImageRGBA->itsPixels,
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

CleanUpSetUpOneTexture:

	//	Free the image (if necessary).
	FreeImageRGBA(&theImageRGBA);

	if (theErrorMessage != NULL)
	{
		//	Substitute a pure red texture for the missing one.
		SetTextureImageFancy(	*aTextureName,
								1,
								1,
								4,
								(Byte [4]){0xFF, 0x00, 0x00, 0xFF},	//	premultiplied alpha
								md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}

static void LoadPaintingTexture(
	ModelData		*md,
	unsigned int	aTextureName,
	unsigned int	aFaceIndex,
	AnaglyphRange	anAnaglyphRange,
	ErrorText		*aFirstError)
{
	ErrorText		theErrorMessage		= NULL;
	ktFacePainting	*thePainting		= NULL;
	char			*theTextureFileName	= NULL;
	ImageRGBA		*theImageRGBA		= NULL;
	unsigned int	theNumPixels		= 0;
	PixelRGBA		*theGreyscalePixels	= NULL;
	unsigned int	i					= 0;

	static char	*theFaceTextureFileNames[] =
				{
					"dummy",
					"MarblePink.rgba",
					"MarbleCyan.rgba",
					"WoodPink.rgba",
					"TopoMap.rgba",
					"Mushroom.rgba",
					"FlowersMagenta.rgba",
					"Sunset.rgba",
					"BlueHills.rgba",
					"Flowers.rgba",
					"Butterflies.rgba"
				};
	static char	*theBackgroundTextureFileNames[] =
				{
					"dummy",
					"WoodYellow.rgba",
					"Clouds.rgba",
					"StoneWhite.rgba",
					"Grass.rgba",
					"Weave.rgba",
					"Moss.rgba",
					"Sand.rgba",
					"Water.rgba",
					"Straw.rgba",
					"Stones.rgba"
				};

	thePainting = &md->itsFacePainting[aFaceIndex];

	switch (thePainting->itsStyle)
	{
		case FacePaintTextureStandard:

			if (aFaceIndex != BACKGROUND_IMAGE)
			{
				if (thePainting->itsStandardTextureID >= 1
				 && thePainting->itsStandardTextureID < BUFFER_LENGTH(theFaceTextureFileNames))
				{
					theTextureFileName = theFaceTextureFileNames[thePainting->itsStandardTextureID];
				}
				else
				{
					theErrorMessage = "Internal Error:  itsStandardTextureID lies beyond the valid range.";
					goto CleanUpLoadPaintingTexture;
				}
			}
			else
			{
				if (thePainting->itsStandardTextureID >= 1
				 && thePainting->itsStandardTextureID < BUFFER_LENGTH(theBackgroundTextureFileNames))
				{
					theTextureFileName = theBackgroundTextureFileNames[thePainting->itsStandardTextureID];
				}
				else
				{
					theErrorMessage = "Internal Error:  itsStandardTextureID lies beyond the valid range.";
					goto CleanUpLoadPaintingTexture;
				}
			}

			//	Read the texture file.
			if ((theErrorMessage = ReadImageRGBA(	theTextureFileName,
													md->itsStereoMode == StereoAnaglyphic
													 && anAnaglyphRange == AnaglyphGrey,
													&theImageRGBA)) != NULL)
				goto CleanUpLoadPaintingTexture;

			//	Pass the pixel data to OpenGL.
			SetTextureImageFancy(	aTextureName,
									theImageRGBA->itsWidth,
									theImageRGBA->itsHeight,
									4,
									(Byte *)theImageRGBA->itsPixels,
									md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

			break;

		case FacePaintTextureCustom:

			if (thePainting->itsCustomTextureSize > 0
			 && thePainting->itsCustomTexturePixels != NULL)
			{
				//	Convert to greyscale if desired.
				if (md->itsStereoMode == StereoAnaglyphic
				 && anAnaglyphRange == AnaglyphGrey)
				{
					theNumPixels		= thePainting->itsCustomTextureSize * thePainting->itsCustomTextureSize;
					theGreyscalePixels	= (PixelRGBA *) GET_MEMORY(theNumPixels * sizeof(PixelRGBA));
					if (theGreyscalePixels != NULL)
					{
						for (i = 0; i < theNumPixels; i++)
							theGreyscalePixels[i] = thePainting->itsCustomTexturePixels[i];

						ConvertPixelsToGreyscale(theGreyscalePixels, theNumPixels);
					}
				}

				//	Pass the pixel data to OpenGL.
				SetTextureImageFancy(	aTextureName,
										thePainting->itsCustomTextureSize,
										thePainting->itsCustomTextureSize,
										4,
										theGreyscalePixels != NULL ?
											(Byte *)theGreyscalePixels :
											(Byte *)thePainting->itsCustomTexturePixels,
										md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
			}
			else
			{
				//	Provide a white texture as a default.
				SetTextureImageFancy(	aTextureName,
										1,
										1,
										4,
										(Byte [4]){0xFF, 0xFF, 0xFF, 0xFF},	//	premultiplied alpha
										md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
			}

			break;

		default:
			//	Ignore this texture -- it's not needed.
			break;
	}

CleanUpLoadPaintingTexture:

	//	Free the image (if present).
	FreeImageRGBA(&theImageRGBA);

	//	Free the greyscale pixels (if present).
	FREE_MEMORY_SAFELY(theGreyscalePixels);

	if (theErrorMessage != NULL)
	{
		//	Substitute a pure red texture for the missing one.
		SetTextureImageFancy(	aTextureName,
								1,
								1,
								4,
								(Byte [4]){0xFF, 0x00, 0x00, 0xFF},	//	premultiplied alpha
								md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}

static void SetCheckerboardTexels(unsigned int aTextureName)
{
	static const Byte	theCheckerboardPixels[2*2*4] =
						{
							0xFF, 0xFF, 0xFF, 0xFF,    0xC0, 0xC0, 0xC0, 0xFF,
							0xC0, 0xC0, 0xC0, 0xFF,    0xFF, 0xFF, 0xFF, 0xFF
						};
	
	glBindTexture(GL_TEXTURE_2D, aTextureName);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 2, 2, 0, GL_RGBA, GL_UNSIGNED_BYTE, theCheckerboardPixels);

	//	For GL_NEAREST to work properly, we must also set GL_TEXTURE_MAX_ANISOTROPY_EXT to 1.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1.0);
}

static ErrorText ReadImageRGBA(
	char		*aTextureFileName,	//	input, zero-terminated UTF-8 string
	bool		aGreyscaleFlag,		//	input
	ImageRGBA	**anImageRGBA)		//	output
{
	ErrorText		theErrorMessage		= NULL;
	ImageRGBA		*theImage			= NULL;
	char			theFilePath[4096]	= "";
	unsigned int	i;
	float			theAlphaFraction;
	
	static bool		theFractionsAreInitialized	= false;
	static float	theFractions[256];
	
	//	Check the output variable.
	if (anImageRGBA == NULL)
	{
		theErrorMessage = "anImageRGBA == NULL in ReadImageRGBA()";
		goto CleanUpReadImageRGBA;
	}
	if (*anImageRGBA != NULL)
	{
		theErrorMessage = "*anImageRGBA != NULL in ReadImageRGBA()";
		goto CleanUpReadImageRGBA;
	}

	//	Allocate and initialize an empty ImageRGBA.
	theImage = GET_MEMORY(sizeof(ImageRGBA));
	if (theImage == NULL)
		goto CleanUpReadImageRGBA;
	theImage->itsNumRawBytes	= 0;
	theImage->itsRawBytes		= NULL;
	theImage->itsWidth			= 0;
	theImage->itsHeight			= 0;
	theImage->itsPixels			= NULL;

	//	Read the texture file's bytes.
	if ((theErrorMessage = GetAbsolutePath("Textures", aTextureFileName, theFilePath, BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpReadImageRGBA;
	if ((theErrorMessage = GetFileContents(theFilePath, &theImage->itsNumRawBytes, &theImage->itsRawBytes)) != NULL)
		goto CleanUpReadImageRGBA;

	//	Did we get a header?
	if (theImage->itsNumRawBytes < 8)
	{
		theErrorMessage = "Texture file lacks width and/or height specification.";
		goto CleanUpReadImageRGBA;
	}

	//	Parse the image width and height as big-endian 4-byte integers.
	theImage->itsWidth  = ((unsigned int)theImage->itsRawBytes[0] << 24)
						+ ((unsigned int)theImage->itsRawBytes[1] << 16)
						+ ((unsigned int)theImage->itsRawBytes[2] <<  8)
						+ ((unsigned int)theImage->itsRawBytes[3] <<  0);
	theImage->itsHeight = ((unsigned int)theImage->itsRawBytes[4] << 24)
						+ ((unsigned int)theImage->itsRawBytes[5] << 16)
						+ ((unsigned int)theImage->itsRawBytes[6] <<  8)
						+ ((unsigned int)theImage->itsRawBytes[7] <<  0);

	//	OpenGL requires power-of-two dimensions.
	if ( ! IsPowerOfTwo(theImage->itsWidth )
	 ||  ! IsPowerOfTwo(theImage->itsHeight) )
	{
		theErrorMessage = "OpenGL requires that each texture's width and height be powers of two.";
		goto CleanUpReadImageRGBA;
	}

	//	Is the total file length correct?
	if (theImage->itsNumRawBytes !=  8  +  4 * theImage->itsWidth * theImage->itsHeight)
	{
		theErrorMessage = "Number of pixels in texture file does not match stated width and height.";
		goto CleanUpReadImageRGBA;
	}

	//	Locate the pixels.
	theImage->itsPixels = (PixelRGBA *) (theImage->itsRawBytes + 8);
	
	//	Premultiply each pixel's red, green and blue components by its alpha component.
	//
	//	Note:  The RGBA files store un-premultiplied values for greater flexibility.
	//	For example, we want to be able to edit an image's alpha mask without losing RGB values.
	if ( ! theFractionsAreInitialized )
	{
		for (i = 0; i < 256; i++)
			theFractions[i] = i / 255.0;
		theFractionsAreInitialized = true;
	}
	for (i = 0; i < theImage->itsWidth * theImage->itsHeight; i++)
	{
		if (theImage->itsPixels[i].a != 0xFF)
		{
			theAlphaFraction = theFractions[theImage->itsPixels[i].a];
			theImage->itsPixels[i].r *= theAlphaFraction;
			theImage->itsPixels[i].g *= theAlphaFraction;
			theImage->itsPixels[i].b *= theAlphaFraction;
		}
	}

	//	Convert to greyscale if desired.
	if (aGreyscaleFlag)
		ConvertPixelsToGreyscale(theImage->itsPixels, theImage->itsWidth * theImage->itsHeight);
	
	//	Assign theImage to the output variable.
	*anImageRGBA = theImage;

CleanUpReadImageRGBA:
	
	if (theErrorMessage != NULL)
	{
		FreeImageRGBA(&theImage);
		if (anImageRGBA != NULL)
			*anImageRGBA = NULL;
	}
	
	return theErrorMessage;
};

static void FreeImageRGBA(ImageRGBA **anImage)
{
	if (anImage != NULL && *anImage != NULL)
	{
		FreeFileContents(&(*anImage)->itsNumRawBytes, &(*anImage)->itsRawBytes);

		FREE_MEMORY_SAFELY(*anImage);
	}
}


static void ShutDownTextures(ModelData *md)
{
	unsigned int	i;

	//	The platform-dependent UI #includes KaleidoTile-Common.h,
	//	but shouldn't know about OpenGL.  To keep things clean,
	//	we've defined itsTextureNames as an array of unsigned int's
	//	rather than an array of GLuint's.
	//	For MacOS X compatibility, though, we need to explicitly cast them
	//	to GLuint's before passing the array to glDeleteTextures().

	//	Delete all textures.
	//	glDeleteTextures() will silently ignore zeros and invalid names.
	glDeleteTextures(NumTextures, (GLuint *) md->itsTextureNames);

	//	Clear all texture names.
	for (i = 0; i < NumTextures; i++)
		md->itsTextureNames[i] = 0;
}
