//	KaleidoTileMatrices.c

#include "KaleidoTileMatrices.h"
#include <math.h>

//	How short must a vector be before we get nervous about normalizing it?
#define ZERO_VECTOR_EPSILON	1e-4

//	When is a sum of unnormalized barycentric coordinates too close to zero?
#define BARYCENTRIC_ERROR_EPSILON	1e-10


static void		ComputeUnscaledInverse(ktMatrix m, ktMatrix mm);


void MatrixTranslation(
	ktMatrix	m,
	ktGeometry	geometry,
	char		axis,	//	axis = 'x' or 'y', but not 'z'
	double		d)		//	d = translation distance
{
	//	Set up a translation through a distance d in the direction
	//	of the given coordinate axis.  The axis must be 0 or 1, because
	//	we can't translate the surface in the direction of axis 2,
	//	which is normal to the surface.

	unsigned int	a,	//	axis in direction of translation
					b;	//	axis in unused direction

	switch (axis)
	{
		case 'x':	a = 0; b = 1; break;
		case 'y':	a = 1; b = 0; break;
		default:	return;
	}

	switch (geometry)
	{
		case GeometrySpherical:

			m[a][a] = cos(d);	m[a][b] = 0.0;	m[a][2] = -sin(d);
			m[b][a] = 0.0;		m[b][b] = 1.0;	m[b][2] = 0.0;
			m[2][a] = sin(d);	m[2][b] = 0.0;	m[2][2] = cos(d);

			break;

		case GeometryEuclidean:

			m[a][a] = 1.0;		m[a][b] = 0.0;	m[a][2] = 0.0;
			m[b][a] = 0.0;		m[b][b] = 1.0;	m[b][2] = 0.0;
			m[2][a] = d;		m[2][b] = 0.0;	m[2][2] = 1.0;

			break;

		case GeometryHyperbolic:

			m[a][a] = cosh(d);	m[a][b] = 0.0;	m[a][2] = sinh(d);
			m[b][a] = 0.0;		m[b][b] = 1.0;	m[b][2] = 0.0;
			m[2][a] = sinh(d);	m[2][b] = 0.0;	m[2][2] = cosh(d);

			break;
	}
}


void MatrixRotation(
	ktMatrix	m,
	double		theta)
{
	m[0][0] = cos(theta);	m[0][1] = sin(theta);	m[0][2] = 0.0;
	m[1][0] = -sin(theta);	m[1][1] = cos(theta);	m[1][2] = 0.0;
	m[2][0] = 0.0;			m[2][1] = 0.0;			m[2][2] = 1.0;
}


void MatrixReflection(
	ktMatrix	m,
	char		axis)	//	axis = 'x' or 'y', but not 'z'
{
	unsigned int	a,	//	axis in direction of translation
					b;	//	axis in unused direction

	switch (axis)
	{
		case 'x':	a = 0; b = 1; break;
		case 'y':	a = 1; b = 0; break;
		default:	return;
	}

	m[a][a] = -1.0;	m[a][b] = 0.0;	m[a][2] = 0.0;
	m[b][a] = 0.0;	m[b][b] = 1.0;	m[b][2] = 0.0;
	m[2][a] = 0.0;	m[2][b] = 0.0;	m[2][2] = 1.0;
}


void MatrixIdentity(
	ktMatrix	m)
{
	m[0][0] = 1.0;		m[0][1] = 0.0;		m[0][2] = 0.0;
	m[1][0] = 0.0;		m[1][1] = 1.0;		m[1][2] = 0.0;
	m[2][0] = 0.0;		m[2][1] = 0.0;		m[2][2] = 1.0;
}


void MatrixProduct(
	CNST ktMatrix	m1,
	CNST ktMatrix	m2,
		 ktMatrix	product)
{
	unsigned int	i,
					j,
					k;
	ktMatrix		temp;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
		{
			temp[i][j] =  0.0;
			for (k = 0; k < 3; k++)
				temp[i][j] += m1[i][k] * m2[k][j];
		}

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			product[i][j] = temp[i][j];
}


void MatrixConjugate(
	CNST ktMatrix	m,
	CNST ktMatrix	a,
		 ktMatrix	Ama,
	CNST ktGeometry	geometry)
{
	ktMatrix	a_inverse,
				temp;

	MatrixInverse(a, a_inverse, geometry);
	MatrixProduct(a_inverse, m, temp);
	MatrixProduct(temp, a, Ama);
}


void MatrixInverse(
	CNST ktMatrix	m,
		 ktMatrix	m_inverse,
	CNST ktGeometry	geometry)
{
	unsigned int	i,
					j;
	ktMatrix		result;
	double			a,
					b,
					c,
					d,
					e,
					f;

	switch (geometry)
	{
		case GeometrySpherical:

			for (i = 0; i < 3; i++)
				for (j = 0; j < 3; j++)
					result[i][j] = m[j][i];

			break;

		case GeometryEuclidean:

			//	A Euclidean isometry factors as an element of O(2)
			//	followed by a translation.
			//
			//		(a  b  0)     (a  b  0) (1  0  0)
			//		(c  d  0)  =  (c  d  0) (0  1  0)
			//		(e  f  1)     (0  0  1) (e  f  1)
			//
			//	Its inverse is
			//
			//		( 1  0  0) (a  c  0)     (   a       c    0 )
			//		( 0  1  0) (b  d  0)  =  (   b       d    0 )
			//		(-e -f  1) (0  0  1)     (-ae-bf  -ce-df  1 )

			{
				a = m[0][0];
				b = m[0][1];
				c = m[1][0];
				d = m[1][1];
				e = m[2][0];
				f = m[2][1];

				result[0][0] = a;        result[0][1] = c;        result[0][2] = 0.0;
				result[1][0] = b;        result[1][1] = d;        result[1][2] = 0.0;
				result[2][0] = -a*e-b*f; result[2][1] = -c*e-d*f; result[2][2] = 1.0;
			}

			break;

		case GeometryHyperbolic:

			for (i = 0; i < 3; i++)
				for (j = 0; j < 3; j++)
					result[i][j] = ((i==2)^(j==2)) ? -m[j][i] : m[j][i];

			break;
	}

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			m_inverse[i][j] = result[i][j];
}


void MatrixCopy(
		ktMatrix	dest,
	CNST ktMatrix	source)
{
	unsigned int	i,
					j;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			dest[i][j] = source[i][j];
}


extern void MatrixSum(
	CNST ktMatrix	m1,
	CNST ktMatrix	m2,
		 ktMatrix	sum)
{
	unsigned int	i,
					j;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			sum[i][j] = m1[i][j] + m2[i][j];
}


void MatrixParallelTransport(
	CNST ktVector	u,
	CNST ktVector	v,
		 ktMatrix	m,
	CNST ktGeometry	aGeometry)
{
	ktMetric	theMetric;
	ktVector	a,
				u1,
				v1;
	ktMatrix	m1,
				m2,
				temp;

	//	Find a matrix m that parallel translates u to v along a geodesic.

	//	Amazingly enough, the same computations work in all three geometries.
	//	The first time you read through the following code, I recommend you make
	//	yourself a sketch illustrating the spherical case.
	//	Then re-read it and make a sketch for the hyperbolic case.
	//	Finally read it once more and make a sketch for the Euclidean case.

	//	Handle the special case that u == v, to avoid special cases later on.
	if (u[0] == v[0]  &&  u[1] == v[1]  &&  u[2] == v[2])
	{
		MatrixIdentity(m);
		return;
	}

	//	It's extremely unlikely any user will ever manange u = -v,
	//	but if one does, we don't want to crash.
	if (u[0] + v[0] == 1.0  &&  u[1] + v[1] == 1.0  &&  u[2] + v[2] == 1.0)
	{
		MatrixIdentity(m);
		return;
	}

	//	Proceed with the general case.

	//	Note the appropriate metric.
	//	In the Euclidean case, use MetricHorizontal.
	theMetric = GetMetric(aGeometry, MetricHorizontal);

	//	u×v gives the axis of rotation.
	VectorCrossProduct(u, v, theMetric, a);
	VectorNormalize(a, theMetric);

	//	Extend (a, u) to an orthonormal basis (a, u, u1)
	//	and    (a, v) to an orthonormal basis (a, v, v1).
	//	Note:  In the Euclidean case a and u1 (resp. u2) have unit length
	//	relative to MetricHorizontal while u (resp. v) has unit length
	//	relative to MetricVertical.
	VectorCrossProduct(a, u, theMetric, u1);
	VectorCrossProduct(a, v, theMetric, v1);

	//	The matrix m should take the basis (u, u1, a) to the basis (v, v1, a).

	//	Let m1 (resp. m2) take the standard basis vectors
	//	to (a, u1, u) (resp. to (a, v1, v)).
	MatrixSetRows(m1, a, u1, u);
	MatrixSetRows(m2, a, v1, v);

	//	m = m1⁻¹ * m2 takes (a, u1, u) to (a, v1, v).
	MatrixInverse(m1, temp, aGeometry);
	MatrixProduct(temp, m2, m);
}


void MatrixSetRows(
		 ktMatrix	m,
	CNST ktVector	a,
	CNST ktVector	b,
	CNST ktVector	c)
{
	unsigned int	i;

	for (i = 0; i < 3; i++)
	{
		m[0][i] = a[i];
		m[1][i] = b[i];
		m[2][i] = c[i];
	}
}


void MatrixToMatrix44(
	CNST ktMatrix	m,
		 double		mm[4][4])
{
	//	Copy a 3x3 ktMatrix to a 4x4 matrix of doubles.
	mm[0][0] = m[0][0];  mm[0][1] = m[0][1];  mm[0][2] = m[0][2];  mm[0][3] = 0.0;
	mm[1][0] = m[1][0];  mm[1][1] = m[1][1];  mm[1][2] = m[1][2];  mm[1][3] = 0.0;
	mm[2][0] = m[2][0];  mm[2][1] = m[2][1];  mm[2][2] = m[2][2];  mm[2][3] = 0.0;
	mm[3][0] = 0.0;      mm[3][1] = 0.0;      mm[3][2] = 0.0;      mm[3][3] = 1.0;
}


void VectorTimesMatrix(
	CNST ktVector	v,
	CNST ktMatrix	m,
		 ktVector	vm)
{
	unsigned int	i,
					j;
	ktVector		temp;

	for (i = 0; i < 3; i++)
	{
		temp[i] = 0.0;
		for (j = 0; j < 3; j++)
			temp[i] += v[j] * m[j][i];
	}

	for (i = 0; i < 3; i++)
		vm[i] = temp[i];
}


void VectorPolarToRectangular(
	CNST double		r,
	CNST double		theta,
		 ktVector	v,
	CNST ktGeometry	geometry)
{
	switch (geometry)
	{
		case GeometrySpherical:
			v[0] = sin(r) * cos(theta);
			v[1] = sin(r) * sin(theta);
			v[2] = cos(r);
			break;

		case GeometryEuclidean:
			v[0] = r * cos(theta);
			v[1] = r * sin(theta);
			v[2] = 1.0;
			break;

		case GeometryHyperbolic:
			v[0] = sinh(r) * cos(theta);
			v[1] = sinh(r) * sin(theta);
			v[2] = cosh(r);
			break;
	}
}


void VectorCrossProduct(
	CNST ktVector	u,
	CNST ktVector	v,
	CNST ktMetric	aMetric,
		 ktVector	uxv)
{
	ktVector		theResult;
	unsigned int	i;

	for (i = 0; i < 3; i++)
		theResult[i] = u[(i+1)%3]*v[(i+2)%3] - u[(i+2)%3]*v[(i+1)%3];

	switch (aMetric)
	{
		case MetricEuclidean:
			//	Leave theResult[2] as it is.
			break;

		case MetricMinkowski:
			theResult[2] = - theResult[2];
			break;

		case MetricHorizontal:
			theResult[2] = 0.0;
			break;

		case MetricVertical:
			//	The MetricVertical case would seem to be ill-defined.
			//	For example, (1,0,0)×(0,1,0) could plausibly take
			//	any value of the form (a,b,1).
			break;
	}

	for (i = 0; i < 3; i++)
		uxv[i] = theResult[i];
}


void VectorNormalize(
		 ktVector	v,
	CNST ktMetric	aMetric)
{
	double	theLength;

	theLength = VectorLength(v, aMetric);

	//	Don't try to normalize the unnormalizable.
	//	We could even flag an error here if desired.
	if (theLength < ZERO_VECTOR_EPSILON)
		theLength = 1.0;

	ConstantTimesVector(1.0/theLength, v, v);
}


double VectorLength(
	CNST ktVector	v,
	CNST ktMetric	aMetric)
{
	return sqrt(fabs(VectorInnerProduct(v, v, aMetric)));
}


double VectorInnerProduct(
	CNST ktVector	u,
	CNST ktVector	v,
	CNST ktMetric	aMetric)
{
	switch (aMetric)
	{
		case MetricEuclidean:
			return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];

		case MetricMinkowski:
			return u[0]*v[0] + u[1]*v[1] - u[2]*v[2];

		case MetricHorizontal:
			return u[0]*v[0] + u[1]*v[1];

		case MetricVertical:
			return u[2]*v[2];

		default:
			return 0.0;	//	The compiler insists on a return value.
	}
}


void VectorAverage(
	CNST ktVector	u,
	CNST ktVector	v,
		 ktVector	avg)
{
	unsigned int	i;

	for (i = 0; i < 3; i++)
		avg[i] = (u[i] + v[i]) / 2.0;
}


void VectorSum(
	CNST ktVector	u,
	CNST ktVector	v,
		 ktVector	sum)
{
	unsigned int	i;

	for (i = 0; i < 3; i++)
		sum[i] = u[i] + v[i];
}


void VectorDifference(
	CNST ktVector	u,
	CNST ktVector	v,
		 ktVector	diff)
{
	unsigned int	i;

	for (i = 0; i < 3; i++)
		diff[i] = u[i] - v[i];
}


void VectorCopy(
		 ktVector	dest,
	CNST ktVector	source)
{
	unsigned int	i;

	for (i = 0; i < 3; i++)
		dest[i] = source[i];
}


void VectorNormal(
	CNST ktVector	v,
	CNST ktGeometry	aGeometry,
		 ktVector	n)
{
	//	On the unit sphere, each point may serve as its own normal vector,
	//	so n = v and we're done.  On the Euclidean and hyperbolic planes,
	//	however, the situation is a little more delicate.  Strictly speaking,
	//	normal vectors aren't vectors, they are 1-forms (but we'll keep calling
	//	them normal vectors anyhow, in spite of the abuse of terminology).
	//	If we write a vector v as a row vector, we may write a normal vector n
	//	as a column vector.  The matrix product v n gives the value
	//	of the normal vector n applied to the ordinary vector v.
	//
	//	The question then is, if we transform a vector v by some matrix m,
	//	( v -> v m ), how must we transform the normal vector?
	//	We want the transformation m to leave the product v n invariant,
	//	so  v n -> (v m)(m⁻¹ n) = v (m m⁻¹) n = v n.  Thus multiplying
	//	v on the right by m corresponds to multiplying n on the left
	//	by m⁻¹.  Alternatively, if we want to think of n as a vector,
	//	then n -> n transpose(m⁻¹).
	//
	//	The two requirements
	//
	//		1.	The normal vector fields on S², E² and H² must be invariant
	//			under the actions of Isom(S²) = O(3), Isom(E²) and
	//			Isom(H²) = O(2,1), respectively.
	//
	//		2.	The normal vector at the north pole (0,0,1) is chosen
	//			to be (0,0,1).
	//
	//	uniquely determine the normal vectors:
	//
	//		A.	In the spherical case, transpose(m⁻¹) = m,
	//			so normal vectors transform along with ordinary vectors.
	//			Thus at each point (x,y,z) the above two requirements
	//			force the normal vector to be (x,y,z) as well.
	//
	//		B.	In the Euclidean case transpose(m⁻¹) always takes
	//			(0,0,1) to (0,0,1), so the normal vector is (0,0,1)
	//			everywhere.
	//
	//		C.	In the hyperbolic case, transpose(m⁻¹) is m
	//			with some extra minus signs thrown in, so when
	//			the position vector (0,0,1) goes to (x,y,z),
	//			the normal vector (0,0,1) goes to (-x,-y,z).

	switch (aGeometry)
	{
		case GeometrySpherical:
			n[0] = v[0];
			n[1] = v[1];
			n[2] = v[2];
			break;

		case GeometryEuclidean:
			n[0] = 0.0;
			n[1] = 0.0;
			n[2] = 1.0;
			break;

		case GeometryHyperbolic:
			n[0] = -v[0];
			n[1] = -v[1];
			n[2] =  v[2];
			break;
	}
}


void ConstantTimesVector(
	CNST double	c,
	CNST ktVector	v,
		 ktVector	cv)
{
	unsigned int	i;

	for (i = 0; i < 3; i++)
		cv[i] = c * v[i];
}


void ConstantTimesMatrix(
	CNST double		c,
	CNST ktMatrix	m,
		 ktMatrix	cm)
{
	unsigned int	i,
					j;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			cm[i][j] = c * m[i][j];
}


ktMetric GetMetric(
	ktGeometry	aGeometry,
	ktMetric	aEuclideanChoice)	//	MetricHorizontal or MetricVertical
{
	//	Return the metric corresponding to the given geometry.
	//	In the Euclidean case the context will require one or the other
	//	of the two possibilities, so the caller must specify which to return.
	switch (aGeometry)
	{
		case GeometrySpherical:		return MetricEuclidean;
		case GeometryEuclidean:		return aEuclideanChoice;
		case GeometryHyperbolic:	return MetricMinkowski;
		default:					return MetricEuclidean;	//	The compiler insists on a return value.
	}
}


void ConvertToBarycentric(
	ktVector	aPoint,					//	input
	ktVector	aBarycentricBasis[3],	//	input
	ktVector	aBarycentricPoint)		//	output
{
	//	Express aPoint relative to the given barycentric basis.
	//	In principle we multiply aPoint times the inverse
	//	of aBarycentricBasisInverse to get aBarycentricPoint.
	//	But given that we need to rescale anyhow (so the barycentric
	//	coordinates will sum to 1) we can save some computational effort
	//	and use an "unscaled inverse".
	//	For even greater efficiency, the latter matrix could be pre-computed
	//	when setting up the initial tiling group, but that hardly seems
	//	necessary given how simple the computation is.

	ktMatrix	theUnscaledInverse;
	double		theSum;

	ComputeUnscaledInverse(aBarycentricBasis, theUnscaledInverse);

	VectorTimesMatrix(aPoint, theUnscaledInverse, aBarycentricPoint);

	theSum = aBarycentricPoint[0]
		   + aBarycentricPoint[1]
		   + aBarycentricPoint[2];

	//	The control panel shows only a limited portion of the tiling,
	//	where all points will (I think!) have theSum > 0.
	//	Nevertheless, it seems prudent to allow for points with theSum <= 0.
	//	We needn't do anything heroic -- mainly just avoid the division by zero
	//	(or division by tiny) -- because in all cases ConvertToBarycentric()
	//	gets followed by ClampToFundamentalTriangle().
	if (theSum > BARYCENTRIC_ERROR_EPSILON)
	{
		aBarycentricPoint[0] /= theSum;
		aBarycentricPoint[1] /= theSum;
		aBarycentricPoint[2] /= theSum;
	}
	else
	{
		aBarycentricPoint[0] = 1.0/3.0;
		aBarycentricPoint[1] = 1.0/3.0;
		aBarycentricPoint[2] = 1.0/3.0;
	}
}


static void ComputeUnscaledInverse(
	ktMatrix	m,	//	input
	ktMatrix	mm)	//	output = det(m)*inverse(m)
{
	//	Compute the transpose of the matrix of cofactors,
	//	but don't bother normalizing by the determinant.

	unsigned int	i,
					j;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			mm[i][j] = m[(j+1)%3][(i+1)%3]*m[(j+2)%3][(i+2)%3]
					 - m[(j+1)%3][(i+2)%3]*m[(j+2)%3][(i+1)%3];
}


void ClampToFundamentalTriangle(ktVector aBarycentricPoint)
{
	//	Clamp to the triangle where all three
	//	barycentric coordinates are nonnegative.

	unsigned int	i;
	double			theSum;

	for (i = 0; i < 3; i++)
		if (aBarycentricPoint[i] < 0.0)
		{
			aBarycentricPoint[i] = 0.0;

			theSum = aBarycentricPoint[(i+1)%3] + aBarycentricPoint[(i+2)%3];
			aBarycentricPoint[(i+1)%3] /= theSum;
			aBarycentricPoint[(i+2)%3] /= theSum;
		}
}
