//	KaleidoTileMatrices.h

#pragma once

//	Trying to mark a matrix passed into a function
//	as "const" raises some thorny issues.  In brief,
//	the C compiler balks at casting a non-const matrix
//	to a const matrix.  For an explanation of why
//	the compiler is right to complain, please see
//	"const to non-const.txt".  I'm not sure what
//	the best solution would be.  For now, let's just
//	replace "const" with "CNST" and then #define it
//	away, so human readers can still see it and grasp
//	its intended meaning, without irritating the compiler.
#define CNST


typedef enum
{
	MetricEuclidean,	//	signature +++
	MetricMinkowski,	//	signature ++-
	MetricHorizontal,	//	signature ++0
	MetricVertical		//	signature 00+
} ktMetric;

typedef enum
{
	GeometrySpherical,
	GeometryEuclidean,
	GeometryHyperbolic
} ktGeometry;

//	In the spherical  case a unit vector v  satisfies <v,v> = +1
//		relative to the standard Euclidean inner product.
//	In the Euclidean  case a unit vector v satsifies v[0] == 1.
//	In the hyperbolic case a unit vector v  satisfies <v,v> = -1
//		relative to the standard Minkowski space inner product,
//		with the z-coordinate (coordinate 2) "timelike".
//	In all three cases the image is centered at the "north pole" (0,0,1).
//
//	A ktVector may also represent a triple of barycentric coordinates,
//	which by default gets normalized as v[0] + v[1] + v[2] == 1.
typedef double	ktVector[3];

//	Rigid motions of each space are represented by 3×3 matrices:
//
//	In the spherical case the matrices are elements of O(3).
//
//	In the Euclidean case the matrices take the following form,
//	which allows them to represent an arbitrary isometry
//	of the plane v[0] == 1.
//
//					( cos t  sin t   0  )
//			(x y 1) (-sin t  cos t   0  )
//					(   a      b     1  )
//
//	In the hyperbolic case the matrices are elements of O(2,1).
typedef double	ktMatrix[3][3];


extern void		MatrixTranslation(ktMatrix m, ktGeometry aGeometry, char axis, double d);
extern void		MatrixRotation(ktMatrix m, double theta);
extern void		MatrixReflection(ktMatrix m, char axis);
extern void		MatrixIdentity(ktMatrix m);
extern void		MatrixProduct(CNST ktMatrix m1, CNST ktMatrix m2, ktMatrix product);
extern void		MatrixConjugate(CNST ktMatrix m, CNST ktMatrix a, ktMatrix Ama, CNST ktGeometry geometry);
extern void		MatrixInverse(CNST ktMatrix m, ktMatrix m_inverse, CNST ktGeometry geometry);
extern void		MatrixCopy(ktMatrix dest, CNST ktMatrix source);
extern void		MatrixSum(CNST ktMatrix m1, CNST ktMatrix m2, ktMatrix sum);
extern void		MatrixParallelTransport(CNST ktVector u, CNST ktVector v, ktMatrix m, CNST ktGeometry aGeometry);
extern void		MatrixSetRows(ktMatrix m, CNST ktVector a, CNST ktVector b, CNST ktVector c);
extern void		MatrixToMatrix44(CNST ktMatrix m, double mm[4][4]);

extern void		VectorTimesMatrix(CNST ktVector v, CNST ktMatrix m, ktVector vm);
extern void		VectorPolarToRectangular(CNST double r, CNST double theta, ktVector v, CNST ktGeometry geometry);
extern void		VectorCrossProduct(CNST ktVector u, CNST ktVector v, CNST ktMetric aMetric, ktVector uxv);
extern void		VectorNormalize(ktVector v, CNST ktMetric aMetric);
extern double	VectorLength(CNST ktVector v, CNST ktMetric aMetric);
extern double	VectorInnerProduct(CNST ktVector u, CNST ktVector v, CNST ktMetric aMetric);
extern void		VectorAverage(CNST ktVector u, CNST ktVector v, ktVector avg);
extern void		VectorSum(CNST ktVector u, CNST ktVector v, ktVector sum);
extern void		VectorDifference(CNST ktVector u, CNST ktVector v, ktVector diff);
extern void		VectorCopy(ktVector dest, CNST ktVector source);
extern void		VectorNormal(CNST ktVector v, CNST ktGeometry aGeometry, ktVector n);

extern void		ConstantTimesVector(CNST double c, CNST ktVector v, ktVector cv);
extern void		ConstantTimesMatrix(CNST double c, CNST ktMatrix m, ktMatrix cm);

extern ktMetric	GetMetric(ktGeometry aGeometry, ktMetric aEuclideanChoice);

extern void		ConvertToBarycentric(ktVector aPoint, ktVector aBarycentricBasis[3], ktVector aBarycentricPoint);
extern void		ClampToFundamentalTriangle(ktVector aBarycentricPoint);
