//	KaleidoTileSimulation.c
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Common.h"

//	How much can the simulation advance in one frame?
#define MAX_FRAME_PERIOD	0.1


static void		UpdateTilingPosition(ModelData *md, double aFramePeriod);
static void		UpdateTriplePointPosition(ModelData *md, double aFramePeriod);
static void		KeepTriplePointCentered(ktVector aTriplePointPosition, ktVector aTriplePointVelocity);


bool SimulationWantsUpdates(
	ModelData	*md)
{
	return md->itsTilingVelocity.x != 0.0
		|| md->itsTilingVelocity.y != 0.0
		|| md->itsTilingVelocity.r != 0.0
		|| md->itsTriplePointVelocity[0] != 0.0
		|| md->itsTriplePointVelocity[1] != 0.0
		|| md->itsTriplePointVelocity[2] != 0.0
		|| md->itsArchimedeanKey != NULL
		|| md->itsRedrawRequestFlag;
}


void SimulationUpdate(
	ModelData	*md,
	double		aFramePeriod,	//	in seconds
	double		aTimeStamp)		//	in seconds, since system startup
{
	//	If some external delay suspends the animation for a few seconds
	//	(for example if the user holds down a menu) we'll receive
	//	a huge frame period.  To avoid a discontinuous jump,
	//	limit the frame period to MAX_FRAME_PERIOD.
	//	This limit should also have the desirable effect of slowing the animation
	//	on systems with humble graphics cards and very slow frame rates.
	if (aFramePeriod > MAX_FRAME_PERIOD)
		aFramePeriod = MAX_FRAME_PERIOD;

	//	Update all types of motion.
	UpdateTilingPosition(md, aFramePeriod);
	UpdateTriplePointPosition(md, aFramePeriod);
	
	//	Terminate display of Archimedean name?
	if (md->itsArchimedeanKey != NULL)
	{
		if (aTimeStamp >= md->itsArchimedeanNameDisplayEndTime)
			md->itsArchimedeanKey = NULL;
	}
	
	//	We may assume that the UI-specific code will redraw
	//	the scene after this call to SimulationUpdate() returns,
	//	so this is a convenient spot to clear itsRedrawRequestFlag.
	md->itsRedrawRequestFlag = false;
}


static void UpdateTilingPosition(
	ModelData	*md,
	double		aFramePeriod)
{
	ktMatrix	thePosition,
				theIncrement;

	//	Convert itsTilingPosition to a matrix by integrating it over unit time.
	IntegrateVelocity(1.0, md->itsTilingPosition, md->itsTiling.itsGeometry, thePosition);

	//	Integrate itsTilingVelocity over aFramePeriod seconds
	//	to get the incremental change to the tiling's position.
	IntegrateVelocity(aFramePeriod, md->itsTilingVelocity, md->itsTiling.itsGeometry, theIncrement);

	//	Post-multiply thePosition by theIncrement.
	MatrixProduct(thePosition, theIncrement, thePosition);

	//	If necessary, pre-multiply by a chirality-preserving group element
	//	to keep the north pole's image from straying too far.
	KeepNorthPoleCentered(thePosition, md->itsTiling.itsGenerators, md->itsTiling.itsGeometry);

	//	Convert the new position back to a velocity.
	DifferentiatePosition(thePosition, md->itsTiling.itsGeometry, &md->itsTilingPosition);
}


void UpdateTriplePointPosition(
	ModelData	*md,
	double		aFramePeriod)
{
	ktVector	theIncrement;

	//	Scale itsTriplePointVelocity in proportion to aFramePeriod
	//	to get the position increment for one frame.
	ConstantTimesVector(aFramePeriod,
						md->itsTriplePointVelocity,
						theIncrement);

	//	Add the increment to itsTriplePointPosition.
	VectorSum(	md->itsTriplePointPosition,
				theIncrement,
				md->itsTriplePointPosition);

	//	Don't let the triple point stray outside the fundamental triangle.
	KeepTriplePointCentered(md->itsTriplePointPosition, md->itsTriplePointVelocity);

	//	The coordinates must sum to 1, so it's easy to eliminate
	//	any possible roundoff error (which is likely to be small
	//	in any case).
	md->itsTriplePointPosition[2] = 1.0 -
		(md->itsTriplePointPosition[0] + md->itsTriplePointPosition[1]);
}


static void KeepTriplePointCentered(
	ktVector	aTriplePointPosition,
	ktVector	aTriplePointVelocity)
{
	//	If the barycentric triple point strays out of the fundamental triangle
	//	[with vertices at (1,0,0), (0,1,0) and (0,0,1)], reflect it back in
	//	and adjust the velocity accordingly.

	ktVector		theNewPosition,
					theNewVelocity;
	unsigned int	i;

	for (i = 0; i < 3; i++)
		if (aTriplePointPosition[i] < 0.0)
		{
			theNewPosition[(i+0)%3] =     - aTriplePointPosition[(i+0)%3];
			theNewPosition[(i+1)%3] = 1.0 - aTriplePointPosition[(i+2)%3];
			theNewPosition[(i+2)%3] = 1.0 - aTriplePointPosition[(i+1)%3];
			VectorCopy(aTriplePointPosition, theNewPosition);

			theNewVelocity[(i+0)%3] = - aTriplePointVelocity[(i+0)%3];
			theNewVelocity[(i+1)%3] = - aTriplePointVelocity[(i+2)%3];
			theNewVelocity[(i+2)%3] = - aTriplePointVelocity[(i+1)%3];
			VectorCopy(aTriplePointVelocity, theNewVelocity);
		}
}


void ChangeTiling(
	ModelData		*md,
	unsigned int	p,
	unsigned int	q,
	unsigned int	r)
{
	if (md->itsTiling.itsPQR[0] != p
	 || md->itsTiling.itsPQR[1] != q
	 || md->itsTiling.itsPQR[2] != r)
	{
		//	Initialize the new tiling
		//	while retaining position, velocity, etc.
		//	as well as the existing Euclidean zoom factor.
		SetUpTiling(&md->itsTiling, p, q, r, md->itsTiling.itsEuclideanZoom);

		//	Play an appropriate sound.
		PlayWAV("thud.wav");
	}

	md->itsRedrawRequestFlag = true;
}


void ChangeStyle(
	ModelData		*md,
	ktTilingStyle	aStyle)
{
	if (md->itsTilingStyle != aStyle)
	{
		//	Set the new style.
		md->itsTilingStyle = aStyle;

		//	Play an appropriate sound.
		PlayWAV("pop.wav");
	}
}


void SetFaceColorHS(
	ktFacePainting	*aFacePainting,
	double			aHue,			//	pass -1.0 to retain previous value
	double			aSaturation,	//	pass -1.0 to retain previous value
	double			aLightness,		//	pass -1.0 to retain previous value
	double			anOpacity)		//	pass -1.0 to retain previous value
{
	//	Make the face colored, with the hue-saturation editing panel active.
	aFacePainting->itsStyle = FacePaintColorHS;

	//	Set the HSLA color components as per the caller's request.

	if (aHue >= 0.0 && aHue <= 1.0)
		aFacePainting->itsHSLA.h = aHue;

	if (aSaturation >= 0.0 && aSaturation <= 1.0)
		aFacePainting->itsHSLA.s = aSaturation;

	if (aLightness >= 0.0 && aLightness <= 1.0)
		aFacePainting->itsHSLA.l = aLightness;

	if (anOpacity >= 0.0 && anOpacity <= 1.0)
		aFacePainting->itsHSLA.a = anOpacity;

	//	Leave aFacePainting->itsStandardTextureID unmodified, 
	//	in case the user later returns to the standard texture panel.
	
	//	Leave aFacePainting->itsCustomTexturePixels unmodified,
	//	in case the user later returns to the custom texture panel.
}

void SetFaceColorLA(
	ktFacePainting	*aFacePainting,
	double			aHue,			//	pass -1.0 to retain previous value
	double			aSaturation,	//	pass -1.0 to retain previous value
	double			aLightness,		//	pass -1.0 to retain previous value
	double			anOpacity)		//	pass -1.0 to retain previous value
{
	//	Make the face colored, with the lightness-opacity editing panel active.
	aFacePainting->itsStyle = FacePaintColorLA;

	//	Set the HSLA color components as per the caller's request.

	if (aHue >= 0.0 && aHue <= 1.0)
		aFacePainting->itsHSLA.h = aHue;

	if (aSaturation >= 0.0 && aSaturation <= 1.0)
		aFacePainting->itsHSLA.s = aSaturation;

	if (aLightness >= 0.0 && aLightness <= 1.0)
		aFacePainting->itsHSLA.l = aLightness;

	if (anOpacity >= 0.0 && anOpacity <= 1.0)
		aFacePainting->itsHSLA.a = anOpacity;

	//	Leave aFacePainting->itsStandardTextureID unmodified, 
	//	in case the user later returns to the standard texture panel.
	
	//	Leave aFacePainting->itsCustomTexturePixels unmodified,
	//	in case the user later returns to the custom texture panel.
}

void SetFaceTextureStandard(
	ktFacePainting	*aFacePainting,
	unsigned int	aStandardTextureID,	//	pass 0 to retain previous value
	bool			*aPreparedTexturesFlag)
{
	//	Make the face textured.
	aFacePainting->itsStyle = FacePaintTextureStandard;

	//	Leave aFacePainting->itsHSLA unmodified, in case the user later
	//	returns to the hue-saturation or lightness-opacity panel.

	//	Assign aStandardTextureID.
	if (aStandardTextureID >= 1  &&  aStandardTextureID <= NUM_STANDARD_TEXTURES)
		aFacePainting->itsStandardTextureID	= aStandardTextureID;
	
	//	Leave aFacePainting->itsCustomTexturePixels unmodified,
	//	in case the user later returns to the custom texture panel.
	
	//	We'll need to reload the textures before drawing the next frame.
	*aPreparedTexturesFlag = false;
}

void SetFaceTextureCustom(
	ktFacePainting	*aFacePainting,
	bool			aBackgroundImageFlag,	//	true for background images, false for face images
	bool			aPanelClickFlag,		//	true for click in paint panel, false for click in selector button
	bool			*aPreparedTexturesFlag)
{
	const char		*theErrorTitle		= NULL,
					*theErrorText		= NULL;
	unsigned int	theSrcWidth			= 0,
					theSrcHeight		= 0,
					theDstRow			= 0,
					theDstCol			= 0,
					theDstSize			= 0,
					theShift			= 0,
					theMask				= 0x00000000,
					theSrcColNumerator	= 0,
					theSrcRowNumerator	= 0,
					theSrcCol0			= 0,
					theSrcCol1			= 0,
					theSrcRow0			= 0,
					theSrcRow1			= 0,
					theColWeight		= 0,
					theRowWeight		= 0,
					theNumRows			= 0,
					theNumCols			= 0,
					theSrcRowOffset		= 0,
					theSrcColOffset		= 0,
					theDstRowOffset		= 0,
					theDstColOffset		= 0,
					i,
					j;
	PixelRGBA		*theNewPixels		= NULL,
					*thePixelArray		= NULL,
					*theSrcPixel		= NULL,
					*theDstPixel		= NULL,
					theSrcPixel00,
					theSrcPixel01,
					theSrcPixel10,
					theSrcPixel11,
					theSrcPixel0,
					theSrcPixel1,
					theSrcPixelFinal;

	//	Make the face custom-textured.
	aFacePainting->itsStyle = FacePaintTextureCustom;

	//	Leave aFacePainting->itsHSLA unmodified, in case the user later
	//	returns to the hue-saturation or lightness-opacity panel.

	//	Leave aFacePainting->itsStandardTextureID unmodified, 
	//	in case the user later returns to the standard texture panel.

	//	Ask the platform-dependent UI to fetch an image from the clipboard/pasteboard.
	if ( ! GetImageToPaste(&theSrcWidth, &theSrcHeight, &thePixelArray) )
	{
		//	The clipboard/pasteboard contains no image.
		//	If itsCustomTexturePixels also contains no image...
		if (aFacePainting->itsCustomTexturePixels == NULL)
		{
			//	...then we'd better explain to the user how this feature works.
			theErrorTitle	= GetLocalizedText("ErrorPaste1Title");
			theErrorText	= GetLocalizedText("ErrorPaste1Message");
		}
		else
		{
			//	itsCustomTexturePixels already contains an image.
			//	If the user clicked on the paint panel itself...
			if (aPanelClickFlag)
			{
				//	...then we'd better explain that the user needs to copy an image first.
				theErrorTitle	= GetLocalizedText("ErrorPaste2Title");
				theErrorText	= GetLocalizedText("ErrorPaste2Message");
			}
			else
			{
				//	The user clicked on the selector button (the little clipboard).
				//	We'll assume s/he wants to return to the previously pasted image.
				//	Clean up and return with no error message.
				theErrorTitle	= NULL;
				theErrorText	= NULL;
			}
		}
		
		//	Whether we're posting an error or not, we're ready to clean up.
		//	Note that the old image, if any, remains in place.
		goto CleanUpSetFaceTextureCustom;
	}

	//	Guard against ridiculous inputs.

	//	This check will ultimately let us multiply theDstSize * theDstSize without fear of overflow.
	if (theSrcWidth  > 0x00004000
	 || theSrcHeight > 0x00004000)
	{
		theErrorTitle	= GetLocalizedText("ErrorImageSizeTitle");
		theErrorText	= GetLocalizedText("ErrorImageSizeMessage");
		goto CleanUpSetFaceTextureCustom;
	}
	
	//	Receiving an empty image is even less likely
	//	than receiving one that's too big,
	//	but subsequent code will be simpler if we can assume
	//	the image's width and height are both nonzero.
	//	Leave this error message in English so as not to bother 
	//	the translators with a message that might never appear.
	if (theSrcWidth  == 0
	 || theSrcHeight == 0)
	{
		theErrorTitle	= "Image is empty";
		theErrorText	= "Image width or height is zero.";
		goto CleanUpSetFaceTextureCustom;
	}
	
	//	Our rule will be to choose the smallest
	//	power-of-two square that accommodates at least the smaller
	//	of the image's two dimensions.  For example, if the image were 300 × 200
	//	we'd use a 256 × 256 square, which accommodates the image height of 200
	//	while omitting portions along the sides.
	//
	//	In the case of a background texture it'll also be handy to have 
	//	pre-computed quantities for quickly computing quotients and remainders
	//	modulo theDstSize.  For example, if theDstSize = 256 = 0x00000100,
	//	then theShift = 8 and theMask = 0x000000FF, and for any unsigned integer n
	//	we may compute n / theDstSize as n >> theShift and
	//	we may compute n % theDstSize as n & theMask.
	//	[Alas in actual practice these "optimizations" made no perceptible difference
	//	to the time required to load an image.  A fairly large image (for example,
	//	a screenshot of KaleidoTile itself) takes a second or two to load on my iMac,
	//	with or without the "optimizations".  So I might want to remove them eventually,
	//	to keep the code conceptually simple.  For now I'll leave them in place.]
	//
	//	Note that the preceding error check guarantee that theSrcWidth and theSrcHeight
	//	are at least 1 and at most 0x4000.
	theDstSize	= 1;
	theShift	= 0;
	theMask		= 0x00000000;
	while (theDstSize < theSrcWidth
		&& theDstSize < theSrcHeight)
	{
		theDstSize	*= 2;
		theShift	+= 1;
		theMask		= (theMask << 1) + 1;
	}
	
	//	Allocate memory for the square power-of-two texture.
	theNewPixels = GET_MEMORY(theDstSize * theDstSize * sizeof(PixelRGBA));
	if (theNewPixels == NULL)
	{
		theErrorTitle	= "Out of Memory";
		theErrorText	= "Couldn't get memory to create square version of custom texture.";
		goto CleanUpSetFaceTextureCustom;
	}

	//	We've got the memory, so we may now safely replace 
	//	the old custom texture with the new custom texture.
	//	Granted, theNewPixels still contain garbage, but
	//	even if we encounter some error and forget to clear
	//	itsCustomTexturePixels, the worst we risk is a garbled image.
	//	(And in practice in the event of an error we would 
	//	clear itsCustomTexturePixels and the user would see
	//	a pure white texture.)
	FREE_MEMORY_SAFELY(aFacePainting->itsCustomTexturePixels);
	aFacePainting->itsCustomTextureSize		= theDstSize;
	aFacePainting->itsCustomTexturePixels	= theNewPixels;
	
	//	Set the new custom texture.
	
	//	We're hoping the custom texture will be square with power-of-two dimensions.
	//	More often, though, the user will paste in a custom texture
	//	with arbitrary dimensions, and we'll need to create a suitable
	//	power-of-two square image.  How we do this will depend on whether
	//	we're preparing an image for a face (where we'll copy out a subimage
	//	pixel-by-pixel) or for a background (where we'll need to resample
	//	to maintain the image's wrappability).
	if (aBackgroundImageFlag)
	{
		//	Prepare a background image.

		//	At render time we'll use the whole texture.
		aFacePainting->itsCustomTextureFraction	= 1.0;	//	ignored
		
		//	Note the aspect ratio.
		aFacePainting->itsCustomTextureAspectRatio	= (double)theSrcWidth / (double)theSrcHeight;
		
		//	Resample the image.
		//
		//	If we were working with a non-wrapping image (an image on a square rather than
		//	an image on a torus) we'd have to account for the fact that the pixels are centered
		//	at half-integer coordinates, and we'd also have to make special arrangements
		//	to accommodate the potential clamping along the boundaries.
		//	Fortunately a wrapping image (an image on a torus) is much easier to work with.
		//	Because it's cyclic, we can mentally shift it so that the pixel centers fall
		//	on integer coordinates.  (Note:  This implies that the final rescaled image
		//	will be offset from the original image by some small amount but that's OK:
		//	the image is cyclic!)
		theDstPixel = aFacePainting->itsCustomTexturePixels;
		for (theDstRow = 0; theDstRow < theDstSize; theDstRow++)
			for (theDstCol = 0; theDstCol < theDstSize; theDstCol++)
			{
				//	Visualize the texture with coordinates running from 0 to 1
				//	in each direction.  Don't worry which way is up and which 
				//	way is down:  it doesn't matter just so we're consistent.
				//	We're at the point (theDstRow/theDstSize, theDstCol/theDstSize).
				//	We want to express this in units of (1/theSrcWidth, 1/theSrcHeight).
				//	At first it seems like a floating point computation might be required,
				//	but with a little thought we can do it all with integer computations instead.
				//	Write the point's coordinates (h, v) as
				//
				//		(	theDstCol / theDstSize,
				//			theDstRow / theDstSize	)
				//
				//	  = (	(theDstCol / theDstSize) (theSrcWidth  / theSrcWidth),
				//			(theDstRow / theDstSize) (theSrcHeight / theSrcHeight)	)
				//
				//	  = (	(theDstCol * theSrcWidth  / theDstSize) / theSrcWidth,
				//			(theDstRow * theSrcHeight / theDstSize) / theSrcHeight	)
				//
				//	A minute of reflection shows that the integer parts of
				//
				//			theDstCol * theSrcWidth  / theDstSize
				//	and
				//			theDstRow * theSrcHeight / theDstSize
				//
				//	give the coordinates of a nearby source pixel,
				//	while the fractional parts of those same quantities
				//	give the relative weights needed to compute a bilinear average
				//	of the four nearest source pixels.
				theSrcColNumerator = theDstCol * theSrcWidth;	//	less than 0x00004000 * 0x00004000 so no danger of overflow
				theSrcRowNumerator = theDstRow * theSrcHeight;	//	less than 0x00004000 * 0x00004000 so no danger of overflow

				theSrcCol0 = theSrcColNumerator >> theShift;	//	= theSrcColNumerator / theDstSize;
				theSrcRow0 = theSrcRowNumerator >> theShift;	//	= theSrcRowNumerator / theDstSize;
				
				theSrcCol1 = theSrcCol0 + 1;
				if (theSrcCol1 == theSrcWidth)	//	realizes the wraparound
					theSrcCol1 = 0;
				
				theSrcRow1 = theSrcRow0 + 1;
				if (theSrcRow1 == theSrcHeight)	//	realizes the wraparound
					theSrcRow1 = 0;
				
				theSrcPixel00 = thePixelArray[theSrcRow0*theSrcWidth + theSrcCol0];
				theSrcPixel10 = thePixelArray[theSrcRow0*theSrcWidth + theSrcCol1];
				theSrcPixel01 = thePixelArray[theSrcRow1*theSrcWidth + theSrcCol0];
				theSrcPixel11 = thePixelArray[theSrcRow1*theSrcWidth + theSrcCol1];

				theColWeight = theSrcColNumerator & theMask;	//	= theSrcColNumerator % theDstSize;
				theRowWeight = theSrcRowNumerator & theMask;	//	= theSrcRowNumerator % theDstSize;
				
				//	Each of the following 12 lines of code realizes " / theDstSize" as " >> theShift".
				
				theSrcPixel0.r = ((theDstSize - theRowWeight) * theSrcPixel00.r + theRowWeight * theSrcPixel01.r) >> theShift;	//	= … / theDstSize;
				theSrcPixel0.g = ((theDstSize - theRowWeight) * theSrcPixel00.g + theRowWeight * theSrcPixel01.g) >> theShift;	//	= … / theDstSize;
				theSrcPixel0.b = ((theDstSize - theRowWeight) * theSrcPixel00.b + theRowWeight * theSrcPixel01.b) >> theShift;	//	= … / theDstSize;
				theSrcPixel0.a = ((theDstSize - theRowWeight) * theSrcPixel00.a + theRowWeight * theSrcPixel01.a) >> theShift;	//	= … / theDstSize;
				
				theSrcPixel1.r = ((theDstSize - theRowWeight) * theSrcPixel10.r + theRowWeight * theSrcPixel11.r) >> theShift;	//	= … / theDstSize;
				theSrcPixel1.g = ((theDstSize - theRowWeight) * theSrcPixel10.g + theRowWeight * theSrcPixel11.g) >> theShift;	//	= … / theDstSize;
				theSrcPixel1.b = ((theDstSize - theRowWeight) * theSrcPixel10.b + theRowWeight * theSrcPixel11.b) >> theShift;	//	= … / theDstSize;
				theSrcPixel1.a = ((theDstSize - theRowWeight) * theSrcPixel10.a + theRowWeight * theSrcPixel11.a) >> theShift;	//	= … / theDstSize;
				
				theSrcPixelFinal.r = ((theDstSize - theColWeight) * theSrcPixel0.r + theColWeight * theSrcPixel1.r) >> theShift;	//	= … / theDstSize;
				theSrcPixelFinal.g = ((theDstSize - theColWeight) * theSrcPixel0.g + theColWeight * theSrcPixel1.g) >> theShift;	//	= … / theDstSize;
				theSrcPixelFinal.b = ((theDstSize - theColWeight) * theSrcPixel0.b + theColWeight * theSrcPixel1.b) >> theShift;	//	= … / theDstSize;
				theSrcPixelFinal.a = ((theDstSize - theColWeight) * theSrcPixel0.a + theColWeight * theSrcPixel1.a) >> theShift;	//	= … / theDstSize;
		
				*theDstPixel++ = theSrcPixelFinal;
			}
	}
	else
	{
		//	Note what fraction of the texture will get filled with images pixels,
		//	so we can adjust our texture coordinates accordingly at render time.
		//	More precisely, what is the size of the largest useful square
		//	centered in the custom texture.
		aFacePainting->itsCustomTextureFraction = (double) MINIMUM(theSrcWidth, theSrcHeight)
												/ (double) theDstSize;
		
		//	The aspect ratio field isn't needed for face textures.
		aFacePainting->itsCustomTextureAspectRatio = 1.0;	//	ignored
		
		//	Copy pixels from the original rectangular image
		//	to the square power-of-two image, to the extent that the two overlap.

		//	Pre-compute row offsets.
		if (theSrcHeight >= theDstSize)
		{
			theNumRows		= theDstSize;
			theSrcRowOffset	= (theSrcHeight - theDstSize) / 2;
			theDstRowOffset	= 0;
		}
		else
		{
			theNumRows		= theSrcHeight;
			theSrcRowOffset	= 0;
			theDstRowOffset	= (theDstSize - theSrcHeight) / 2;
		}

		//	Pre-compute column offsets.
		if (theSrcWidth >= theDstSize)
		{
			theNumCols		= theDstSize;
			theSrcColOffset	= (theSrcWidth - theDstSize) / 2;
			theDstColOffset	= 0;
		}
		else
		{
			theNumCols		= theSrcWidth;
			theSrcColOffset	= 0;
			theDstColOffset	= (theDstSize - theSrcWidth) / 2;
		}
		
		//	Copy the pixels.
		for (i = 0; i < theNumRows; i++)
		{
			theSrcPixel = &thePixelArray[(theSrcRowOffset + i)*theSrcWidth + theSrcColOffset];
			theDstPixel = &aFacePainting->itsCustomTexturePixels[(theDstRowOffset + i)*theDstSize + theDstColOffset];

			for (j = theNumCols; j-- > 0; )
				*theDstPixel++ = *theSrcPixel++;
		}
		
		//	If the source image doesn't completely fill the destination image,
		//	duplicate the boundary pixels so that at render time, interpolation
		//	will give a reasonable result.
		//	We'll need more than a 1-pixel border zone to allow for mipmapping,
		//	so copy the boundary pixels all the way to the edge of the destination bitmap.
		//
		//	Technical Note:  It's easier to copy whole rows and columns than
		//	to bother figuring out where the minimum essential sub-rows and sub-columns lie.
		if (theDstColOffset > 0)
		{
			for (theDstCol = 0; theDstCol < theDstColOffset; theDstCol++)
				for (theDstRow = 0; theDstRow < theDstSize; theDstRow++)
					  aFacePainting->itsCustomTexturePixels[theDstRow*theDstSize + theDstCol      ]
					= aFacePainting->itsCustomTexturePixels[theDstRow*theDstSize + theDstColOffset];
		}
		if (theDstColOffset + theNumCols < theDstSize)
		{
			for (theDstCol = theDstColOffset + theNumCols; theDstCol < theDstSize; theDstCol++)
				for (theDstRow = 0; theDstRow < theDstSize; theDstRow++)
					  aFacePainting->itsCustomTexturePixels[theDstRow*theDstSize + theDstCol                       ]
					= aFacePainting->itsCustomTexturePixels[theDstRow*theDstSize + theDstColOffset + theNumCols - 1];
		}
		if (theDstRowOffset > 0)
		{
			for (theDstRow = 0; theDstRow < theDstRowOffset; theDstRow++)
				for (theDstCol = 0; theDstCol < theDstSize; theDstCol++)
					  aFacePainting->itsCustomTexturePixels[   theDstRow   *theDstSize + theDstCol]
					= aFacePainting->itsCustomTexturePixels[theDstRowOffset*theDstSize + theDstCol];
		}
		if (theDstRowOffset + theNumRows < theDstSize)
		{
			for (theDstRow = theDstRowOffset + theNumRows; theDstRow < theDstSize; theDstRow++)
				for (theDstCol = 0; theDstCol < theDstSize; theDstCol++)
					  aFacePainting->itsCustomTexturePixels[            theDstRow             *theDstSize + theDstCol]
					= aFacePainting->itsCustomTexturePixels[(theDstRowOffset + theNumRows - 1)*theDstSize + theDstCol];
		}
	}

CleanUpSetFaceTextureCustom:

	//	Let the UI free aPixelArray.
	FreeImageToPaste(&thePixelArray);
	
	//	We'll need to reload the textures before drawing the next frame.
	*aPreparedTexturesFlag = false;

	//	If an error occurred, deal the error message.
	if (theErrorTitle != NULL && theErrorText != NULL)
		ErrorMessage(theErrorText, theErrorTitle);
}

void SetFaceInvisible(ktFacePainting *aFacePainting)
{
	//	Make the face invisible.
	aFacePainting->itsStyle = FacePaintInvisible;

	//	Leave aFacePainting->itsHSLA unmodified, in case the user later
	//	returns to the hue-saturation or lightness-opacity panel.
	
	//	Leave aFacePainting->itsStandardTextureID unmodified, 
	//	in case the user later returns to the standard texture panel.
	
	//	Leave aFacePainting->itsCustomTexturePixels unmodified,
	//	in case the user later returns to the custom texture panel.
}

