//	KaleidoTileTiling.c

#include <math.h>
#include "KaleidoTile-Common.h"
#include "KaleidoTileTiling.h"


//	By default, how big should the tiles be in a Euclidean tiling?
#define DEFAULT_EUCLIDEAN_INRADIUS	0.1

//	Two matrices are considered equal iff they take the north pole (0,0,1)
//	to the same vector, to an accuracy of MATRIX_EPSILON in each coordinate.
//	MATRIX_EPSILON can be large because the group is discrete and
//	the north pole (0,0,1) lies well within the triangle's interior.
#define MATRIX_EPSILON	1e-3

//	Guard against infinite loops when keeping the triple point near the north pole.
#define VECTOR_COMPARISON_EPSILON	1e-6


static void	InitGeometry(ktTiling *aTiling);
static void	InitBaseTriangle(ktTiling *aTiling);
static void	InitBarycentricBasis(ktTiling *aTiling);
static bool	CandidateVectorIsCloser(ktVector aCandidate, ktVector anIncumbent, ktGeometry aGeometry);


void SetUpTiling(
	ktTiling		*aTiling,
	unsigned int	p,
	unsigned int	q,
	unsigned int	r,
	signed int		aEuclideanZoom)	//	exponent of √2;  always present, but used only for Euclidean tilings
{
	//	Clamp aEuclideanZoom to a reasonable range.
	//	In particular, don't let it get so small that our tiling code
	//	will exhaust it fixed suppy of PerimeterSegments before covering
	//	the whole window.
	//	(The UI should disable the menu commands to Make Tiles Bigger or Smaller
	//	when the user reaches the limit of the allowable range, but let's check
	//	aEuclideanZoom again here just to be safe.)
	if (aEuclideanZoom < MIN_EUCLIDEAN_ZOOM)
		aEuclideanZoom = MIN_EUCLIDEAN_ZOOM;
	if (aEuclideanZoom > MAX_EUCLIDEAN_ZOOM)
		aEuclideanZoom = MAX_EUCLIDEAN_ZOOM;

	//	Record the requested values of (p,q,r).
	aTiling->itsPQR[0] = p;
	aTiling->itsPQR[1] = q;
	aTiling->itsPQR[2] = r;
	
	//	Record the requested Euclidean zoom factor,
	//	for immediate use if the tiling is Euclidean,
	//	or for possible future use otherwise.
	aTiling->itsEuclideanZoom = aEuclideanZoom;

	//	Is this a spherical, Euclidean or hyperbolic tiling?
	InitGeometry(aTiling);

	//	Conceptually the tiling consists of a base triangle,
	//	which will get translated around by the elements
	//	of a (p,q,r) triangle group.  The base triangle
	//	gets cut into three differently-colored quadrilaterals,
	//	which meet at the so-called "triple point".
	//	Compute the triangle's vertices, the matrices describing
	//	the reflections in its sides, and its outradius.
	InitBaseTriangle(aTiling);

	//	Rescale itsVertices to realize the barycentric basis,
	//	relative to which -- up to rescaling --
	//	(1,0,0), (0,1,0) and (0,0,1) are the fundamental triangle's vertices;
	//	(0, 1/2, 1/2), (1/2, 0, 1/2) and (1/2, 1/2, 0) are the intersections
	//	of angle bisectors with opposite sides; and (1/3, 1/3, 1/3) is
	//	the north pole.
	InitBarycentricBasis(aTiling);
}


static void InitGeometry(ktTiling *aTiling)
{
	//	The geometry will be spherical, Euclidean or hyperbolic
	//	according to whether the angle sum is greater than,
	//	equal to, or less than π.
	//
	//			π/p + π/q + π/r  ⋛  π
	//
	//	Multiply through by pqr/π to convert to an integer equation
	//
	//			qr + pr + pq  ⋛  pqr
	//
	//	(p, q and r will be small enough that we needn't worry
	//	about overflows.)

	signed int	theSum;

	theSum = aTiling->itsPQR[1] * aTiling->itsPQR[2]
		   + aTiling->itsPQR[0] * aTiling->itsPQR[2]
		   + aTiling->itsPQR[0] * aTiling->itsPQR[1]
		   - aTiling->itsPQR[0] * aTiling->itsPQR[1] * aTiling->itsPQR[2];

	if (theSum > 0)
		aTiling->itsGeometry = GeometrySpherical;
	else if (theSum == 0)
		aTiling->itsGeometry = GeometryEuclidean;
	else
		aTiling->itsGeometry = GeometryHyperbolic;
}


static void InitBaseTriangle(ktTiling *aTiling)
{
	unsigned int	*pqr,
					i;
	double			theSides[3],
					theTangents[3],
					theThetas[3],
					theRadials[3];
	double			theBoundingBoxLeft,
					theBoundingBoxBottom,
					theBoundingBoxRight,
					theBoundingBoxTop;
	ktMatrix		theReflection,
					theTranslation,
					theRotation;

	//	Before reading the following explanation, draw yourself
	//	a figure as follows.  Begin with an arbitrary triangle and label
	//	its angles π/p, π/q and π/r.  Draw the inscribed circle.
	//	Draw segments from each of the triangle's vertices to the center
	//	of the inscribed circle.  By symmetry, these are the angle
	//	bisectors.  Now draw segments from each point of tangency
	//	(where the inscribed circle touches the triangle) to the
	//	center of the inscribed circle.  These segments meet the
	//	triangle at right angles.  The segments you've drawn divide
	//	the triangle into six smaller triangles.  Focus your attention
	//	on one of the small triangles incident to the vertex of angle π/p.
	//	Label the small triangle's angles π/2p, π/2 and θ_p.

	//	For brevity.
	pqr = aTiling->itsPQR;

	//	The base triangle has angles π/p, π/q and π/r.
	//	Break into cases to analyze it.
	//	The spherical and hyperbolic cases are almost identical.
	//	The Euclidean case is exceptional.
	switch (aTiling->itsGeometry)
	{
		case GeometrySpherical:

			//	First find the lengths of the big triangle's sides.
			//	Use the spherical law of cosines for angles to solve
			//	for the length of side c in terms of the angles A,B,C.
			//
			//		cos C = -cos A cos B + sin A sin B cos c
			//
			for (i = 0; i < 3; i++)
				theSides[i] = acos(
					(cos(PI/pqr[i]) + cos(PI/pqr[(i+1)%3])*cos(PI/pqr[(i+2)%3]))
					/ (sin(PI/pqr[(i+1)%3])*sin(PI/pqr[(i+2)%3])));

			//	Let theTangents[i] be the length of the segment from
			//	vertex i to either point of tangency where the inscribed
			//	circle touches the triangle's side.  Clearly
			//
			//		theSides[i] = theTangents[(i+1)%3] + theTangents[(i+2)%3]
			//
			//	so it's easy to solve for theTangents.
			for (i = 0; i < 3; i++)
				theTangents[i] = (theSides[(i+1)%3] + theSides[(i+2)%3] - theSides[i]) / 2.0;

			//	Now consider each of the six small triangles described
			//	in the documentation at the beginning of this function.
			//	Apply the spherical law of cosines for angles to solve
			//	for θ_p, θ_q, and θ_r.
			for (i = 0; i < 3; i++)
				theThetas[i] = acos( cos(theTangents[i]) * sin(PI/(2.0*pqr[i])) );

			//	Apply the spherical law of cosines for angles to solve
			//	for the inradius.
			aTiling->itsInradius = acos( cos(PI/(2.0*pqr[0])) / sin(theThetas[0]) );

			//	Apply the spherical law of cosines for angles once more
			//	to solve for the radial distance from the incenter to
			//	each vertex of the large triangle.
			for (i = 0; i < 3; i++)
				theRadials[i] = acos( 1.0 / (tan(PI/(2.0*pqr[i])) * tan(theThetas[i])) );

			break;

		case GeometryEuclidean:

			//	The θ's are the complements of π/2p, π/2q and π/2r.
			for (i = 0; i < 3; i++)
				theThetas[i] = (PI/2)*(1.0 - 1.0/pqr[i]);

			//	In the Euclidean case, the fundamental triangle's inradius
			//	may be anything the user wants.
			aTiling->itsInradius = DEFAULT_EUCLIDEAN_INRADIUS * pow(sqrt(2.0), aTiling->itsEuclideanZoom);

			//	Solve for theRadials.
			for (i = 0; i < 3; i++)
				theRadials[i] = aTiling->itsInradius / sin(PI/(2.0*pqr[i]));

			break;

		case GeometryHyperbolic:

			//	The hyperbolic case is identical to the spherical case,
			//	except that cosh() and acosh() sometimes appear in place
			//	of cos() and acos().

			//	First find the lengths of the big triangle's sides.
			//	Use the hyperbolic law of cosines for angles to solve
			//	for the length of side c in terms of the angles A,B,C.
			//
			//		cos C = -cos A cos B + sin A sin B cosh c
			//
			for (i = 0; i < 3; i++)
				theSides[i] = acosh(
					(cos(PI/pqr[i]) + cos(PI/pqr[(i+1)%3])*cos(PI/pqr[(i+2)%3]))
					/ (sin(PI/pqr[(i+1)%3])*sin(PI/pqr[(i+2)%3])));

			//	Let theTangents[i] be the length of the segment from
			//	vertex i to either point of tangency where the inscribed
			//	circle touches the triangle's side.  Clearly
			//
			//		theSides[i] = theTangents[(i+1)%3] + theTangents[(i+2)%3]
			//
			//	so it's easy to solve for theTangents.
			for (i = 0; i < 3; i++)
				theTangents[i] = (theSides[(i+1)%3] + theSides[(i+2)%3] - theSides[i]) / 2.0;

			//	Now consider each of the six small triangles described
			//	in the documentation at the beginning of this function.
			//	Apply the hyperbolic law of cosines for angles to solve
			//	for θ_p, θ_q, and θ_r.
			for (i = 0; i < 3; i++)
				theThetas[i] = acos( cosh(theTangents[i]) * sin(PI/(2.0*pqr[i])) );

			//	Apply the hyperbolic law of cosines for angles to solve
			//	for the inradius.
			aTiling->itsInradius = acosh( cos(PI/(2.0*pqr[0])) / sin(theThetas[0]) );

			//	Apply the hyperbolic law of cosines for angles once more
			//	to solve for the radial distance from the incenter to
			//	each vertex of the large triangle.
			for (i = 0; i < 3; i++)
				theRadials[i] = acosh( 1.0 / (tan(PI/(2.0*pqr[i])) * tan(theThetas[i])) );

			break;
	}

	//	Compute the reflection matrix for each side of the base triangle.
	//	Position the incenter at the north pole (0,0,1).
	//	The plan is to start with a reflection in the direction of the y-axis,
	//	conjugate it by a translation to get itsGenerators[0], and then
	//	conjugate by rotations to get itsGenerators[1] and itsGenerators[2].

	//	theReflection is	1  0  0
	//						0 -1  0
	//						0  0  1
	MatrixReflection(theReflection, 'y');

	//	theTranslation is a translation in the direction of the
	//	negative y-axis, through a distance equal to itsInradius.
	MatrixTranslation(theTranslation, aTiling->itsGeometry, 'y', - aTiling->itsInradius);

	//	Conjugate theReflection by theTranslation to get itsGenerators[0].
	MatrixConjugate(theReflection, theTranslation, aTiling->itsGenerators[0], aTiling->itsGeometry);

	//	Rotate itsGenerators[0] about the north pole (0,0,1)
	//	to get itsGenerators[1] and itsGenerators[2].
	//	A sketch may help here.
	//	Viewed from above the north pole, the vertices of angle
	//	π/p, π/q and π/r go in counterclockwise order.

	MatrixRotation(theRotation, 2*theThetas[2]);
	MatrixConjugate(aTiling->itsGenerators[0], theRotation, aTiling->itsGenerators[1], aTiling->itsGeometry);

	MatrixRotation(theRotation, -2*theThetas[1]);
	MatrixConjugate(aTiling->itsGenerators[0], theRotation, aTiling->itsGenerators[2], aTiling->itsGeometry);


	//	itsDomainRadius is the maximum distance from the center to a vertex.
	aTiling->itsDomainRadius = 0.0;
	for (i = 0; i < 3; i++)
		if (aTiling->itsDomainRadius < theRadials[i])
			aTiling->itsDomainRadius = theRadials[i];


	//	Compute the vertex coordinates in rectangular coordinates.
	VectorPolarToRectangular(	theRadials[0],
								-PI/2 + 2*theThetas[2] + theThetas[0],
								aTiling->itsVertices[0],
								aTiling->itsGeometry);
	VectorPolarToRectangular(	theRadials[1],
								-PI/2 - theThetas[1],
								aTiling->itsVertices[1],
								aTiling->itsGeometry);
	VectorPolarToRectangular(	theRadials[2],
								-PI/2 + theThetas[2],
								aTiling->itsVertices[2],
								aTiling->itsGeometry);

	//	When rendering a single fundamental domain we'll want to know
	//	how to center it (itsDomainOffset) and how big it is (itsDomainScale).
	//	In the spherical case the scaled vertices will be closer to the origin,
	//	which isn't a problem;  in the hyperbolic case the scaled vertices
	//	will be further from the origin, which means they might get clipped
	//	a bit, but I'm hoping this won't be a big deal for low order groups.
	theBoundingBoxLeft		= 0.0;
	theBoundingBoxBottom	= 0.0;
	theBoundingBoxRight		= 0.0;
	theBoundingBoxTop		= 0.0;
	for (i = 0; i < 3; i++)
	{
		if (theBoundingBoxLeft		> aTiling->itsVertices[i][0])
			theBoundingBoxLeft		= aTiling->itsVertices[i][0];
		if (theBoundingBoxBottom	> aTiling->itsVertices[i][1])
			theBoundingBoxBottom	= aTiling->itsVertices[i][1];
		if (theBoundingBoxRight		< aTiling->itsVertices[i][0])
			theBoundingBoxRight		= aTiling->itsVertices[i][0];
		if (theBoundingBoxTop		< aTiling->itsVertices[i][1])
			theBoundingBoxTop		= aTiling->itsVertices[i][1];
	}
	aTiling->itsDomainOffset = -(theBoundingBoxLeft + theBoundingBoxRight)/2.0;
	aTiling->itsDomainScale = (theBoundingBoxRight - theBoundingBoxLeft)/2.0;
	if (aTiling->itsDomainScale < -theBoundingBoxBottom)
		aTiling->itsDomainScale = -theBoundingBoxBottom;
	if (aTiling->itsDomainScale <  theBoundingBoxTop)
		aTiling->itsDomainScale =  theBoundingBoxTop;
}


static void InitBarycentricBasis(ktTiling *aTiling)
{
	//	Our goal here is to modify the usual barycentric coordinates
	//	for the fundamental triangle so that the "special points",
	//	where each angular bisector crosses the triangle's opposite side,
	//	have barycentric coordinates 1/2, 1/2 and 0 (in some order) and
	//	of course the incenter has barycentric coordinates (1/3, 1/3, 1/3).
	//
	//	Let the triangle's three original vertices be
	//	u = {ux,uy,uz}, v = {vx,vy,vz} and w = {wx,wy,wz},
	//	and search for constants a, b and c such that
	//	the midpoint between a u and b v lies in the plane
	//	spanned by {0,0,0}, {0,0,1} and w, and similarly
	//	for the midpoints of the other two pairs.
	//
	//	The coplanarity condition just mentioned amounts to saying
	//	that the midpoint between a u and b v has the same ratio
	//	of its x and y coordinates as does w.  Writing out the ratios
	//	gives three linear equations in three variables.
	//	One could easily solve these equations by hand, but to save
	//	time I did the computation in the Mathematica notebook
	//	"Nice Barycentric Coordinates.nb", obtaining
	//
	//		a → vx wy - vy wx,
	//		b → wx uy - wy ux,
	//		c → ux vy - uy vx
	//
	//	Note that any consistent scalar multiple of a, b and c
	//	will work equally well.  I suspect the present formulas
	//	will yield positive a, b and c whenever the vertices u, v and w
	//	wind counterclockwise around the north pole (0,0,1).

	unsigned int	i;

	for (i = 0; i < 3; i++)
		ConstantTimesVector(
				aTiling->itsVertices[(i+1)%3][0] * aTiling->itsVertices[(i+2)%3][1]
			  - aTiling->itsVertices[(i+1)%3][1] * aTiling->itsVertices[(i+2)%3][0],
			aTiling->itsVertices[i],
			aTiling->itsBarycentricBasis[i]);
}


void KeepNorthPoleCentered(
	ktMatrix	aPosition,
	ktMatrix	aGeneratorSet[3],
	ktGeometry	aGeometry)
{
	//	KaleidoTile 3 stores itsTilingPosition as a Lie derivative,
	//	which always preserves chirality.  Therefore we may modify
	//	aPosition by pairs of generators, but not by single generators,
	//	because the latter reverse chirality.

	unsigned int	i,
					j;
	ktMatrix		theProtoCandidate,
					theCandidate;

	for (i = 0; i < 3; i++)
	{
		MatrixProduct(aGeneratorSet[i], aPosition, theProtoCandidate);

		for (j = 0; j < 3; j++)
		{
			if (j == i)
				continue;

			MatrixProduct(aGeneratorSet[j], theProtoCandidate, theCandidate);

			if (CandidateVectorIsCloser(theCandidate[2], aPosition[2], aGeometry))
				MatrixCopy(aPosition, theCandidate);
		}
	}
}


static bool CandidateVectorIsCloser(
	ktVector	aCandidate,
	ktVector	anIncumbent,
	ktGeometry	aGeometry)
{
	switch (aGeometry)
	{
		case GeometrySpherical:
			return (aCandidate[2] > anIncumbent[2] + VECTOR_COMPARISON_EPSILON);

		case GeometryEuclidean:
			return (aCandidate [0]*aCandidate [0] +  aCandidate [1]*aCandidate [1]
				  < anIncumbent[0]*anIncumbent[0] +  anIncumbent[1]*anIncumbent[1]
				  - VECTOR_COMPARISON_EPSILON);

		case GeometryHyperbolic:
			return (aCandidate[2] < anIncumbent[2] - VECTOR_COMPARISON_EPSILON);

		default:	//	keep the compiler happy
			return false;
	}
}


bool TilingIs22n(ktTiling *aTiling)
{
	return (aTiling->itsPQR[0] == 2 ? 1 : 0)
		 + (aTiling->itsPQR[1] == 2 ? 1 : 0)
		 + (aTiling->itsPQR[2] == 2 ? 1 : 0)
		 >= 2;
}
