//	KaleidoTileVelocities.c

#include <math.h>
#include "KaleidoTileVelocities.h"
#include "KaleidoTileMatrices.h"

#define SMALL_L_LIMIT	0.001


void IntegrateVelocity(
	double		aTimeInterval,	//	input
	ktVelocity	aVelocity,		//	input
	ktGeometry	aGeometry,		//	input
	ktMatrix	aMotion)		//	output
{
	ktMatrix	m;
	double		theLength,
				theLengthSquared,
				c1,
				c2;
	ktMatrix	theTerm0,
				theTerm1,
				theTerm2;

	switch (aGeometry)
	{
		case GeometrySpherical:
		
			//	In the spherical case, aVelocity represents the matrix
			//
			//			( 0   vr -vx )
			//		m = (-vr  0  -vy )
			//			( vx  vy  0  )
			//
			//	Miraculously, m³ = -(vx² + vy² + vr²)m,
			//	so the nominal infinite series for
			//
			//		exp(m) = 1 + m + m²/(2!) + m³/(3!) + ...
			//
			//	compresses to a quadratic polynomial in m,
			//
			//		exp(m) = 1 + (...)m + (...)m²
			//
			//	Letting L = sqrt(vx² + vy² + vr²),
			//
			//		exp(t m) = 1 + t m + t²m²/(2!) + t³m³/(3!) + ...
			//
			//				 = 1 + (t - t³L²/(3!) + ...) m + (t²/(2!) - ...)m²
			//
			//				 = 1  +  sin(t L) (m/L)  +  (1 - cos(t L)) (m²/L²)
			//
			//	As L goes to zero, the polynomial approaches
			//
			//		exp(t m) = 1 + t m + t²/2 m²
			//
			//	(keeping terms linear in L while discarding terms in L² and higher).

			m[0][0] =  0.0;          m[0][1] =  aVelocity.r;  m[0][2] = -aVelocity.x;
			m[1][0] = -aVelocity.r;  m[1][1] =  0.0;          m[1][2] = -aVelocity.y;
			m[2][0] =  aVelocity.x;  m[2][1] =  aVelocity.y;  m[2][2] =  0.0;

			theLength = sqrt(aVelocity.x * aVelocity.x
						   + aVelocity.y * aVelocity.y
						   + aVelocity.r * aVelocity.r);
			if (theLength < SMALL_L_LIMIT)
			{
				c1 = aTimeInterval;
				c2 = aTimeInterval * aTimeInterval / 2.0;
			}
			else
			{
				c1 = sin(aTimeInterval * theLength) / theLength;
				c2 = (1.0 - cos(aTimeInterval * theLength)) / (theLength * theLength);
			}
			
			break;

		case GeometryEuclidean:

			//	In the Euclidean case, aVelocity represents the matrix
			//
			//			( 0   vr  0 )
			//		m = (-vr  0   0 )
			//			( vx  vy  0 )
			//
			//	Miraculously, m³ = -vr² m,
			//	so the nominal infinite series for
			//
			//		exp(m) = 1 + m + m²/(2!) + m³/(3!) + ...
			//
			//	compresses to a quadratic polynomial in m,
			//
			//		exp(m) = 1 + (...)m + (...)m²
			//
			//	Letting L = |vr|,
			//
			//		exp(t m) = 1 + t m + t²m²/(2!) + t³m³/(3!) + ...
			//
			//				 = 1 + (t - t³L²/(3!) + ...) m + (t²/(2!) - ...)m²
			//
			//				 = 1  +  sin(t L) (m/L)  +  (1 - cos(t L)) (m²/L²)
			//
			//	As L goes to zero, the polynomial approaches
			//
			//		exp(t m) = 1 + t m + t/2 m²
			//
			//	(keeping terms linear in L while discarding terms in L and higher).

			m[0][0] =  0.0;          m[0][1] =  aVelocity.r;  m[0][2] =  0.0;
			m[1][0] = -aVelocity.r;  m[1][1] =  0.0;          m[1][2] =  0.0;
			m[2][0] =  aVelocity.x;  m[2][1] =  aVelocity.y;  m[2][2] =  0.0;

			theLength = fabs(aVelocity.r);
			if (theLength < SMALL_L_LIMIT)
			{
				c1 = aTimeInterval;
				c2 = aTimeInterval * aTimeInterval / 2.0;
			}
			else
			{
				c1 = sin(aTimeInterval * theLength) / theLength;
				c2 = (1.0 - cos(aTimeInterval * theLength)) / (theLength * theLength);
			}
			
			break;

		case GeometryHyperbolic:

			//	In the hyperbolic case, aVelocity represents the matrix
			//
			//			( 0   vr  vx )
			//		m = (-vr  0   vy )
			//			( vx  vy  0  )
			//
			//	Miraculously, m³ = -(-vx² - vy² + vr²)m,
			//	so the nominal infinite series for
			//
			//		exp(m) = 1 + m + m²/(2!) + m³/(3!) + ...
			//
			//	compresses to a quadratic polynomial in m,
			//
			//		exp(m) = 1 + (...)m + (...)m²
			//
			//	Subcase 1.  -vx² - vy² + vr² > 0 (elliptic   motion of the hyperbolic plane).
			//		Let L = sqrt(-vx² - vy² + vr²).
			//
			//	Subcase 2.  -vx² - vy² + vr² = 0 (parabolic  motion of the hyperbolic plane).
			//		Let L = 0.
			//
			//	Subcase 3.  -vx² - vy² + vr² < 0 (hyperbolic motion of the hyperbolic plane).
			//		Let L = i sqrt( vx² + vy² - vr²).
			//
			//		exp(t m) = 1 + t m + t²m²/(2!) + t³m³/(3!) + ...
			//
			//				 = 1 + (t - t³L²/(3!) + ...) m + (t²/(2!) - ...)m²
			//
			//				 = 1  +  sin(t L) (m/L)  +  (1 - cos(t L)) (m²/L²)
			//
			//	As L goes to zero, the polynomial approaches
			//
			//		exp(t m) = 1 + t m + t²/2 m²
			//
			//	(keeping terms linear in L while discarding terms in L and higher).

			m[0][0] =  0.0;          m[0][1] =  aVelocity.r;  m[0][2] =  aVelocity.x;
			m[1][0] = -aVelocity.r;  m[1][1] =  0.0;          m[1][2] =  aVelocity.y;
			m[2][0] =  aVelocity.x;  m[2][1] =  aVelocity.y;  m[2][2] =  0.0;

			theLengthSquared = - aVelocity.x * aVelocity.x
							   - aVelocity.y * aVelocity.y
							   + aVelocity.r * aVelocity.r;
			if (theLengthSquared > SMALL_L_LIMIT*SMALL_L_LIMIT)
			{
				theLength = sqrt(theLengthSquared);
				c1 = sin(aTimeInterval * theLength) / theLength;
				c2 = (1.0 - cos(aTimeInterval * theLength)) / (theLength * theLength);
			}
			else if (theLengthSquared < -SMALL_L_LIMIT*SMALL_L_LIMIT)
			{
				//	If L' = |L|, then L = i L' and
				//
				//		exp(t m) = 1 + (t + t³L'²/(3!) + ...) m + (t²/(2!) + ...)m²
				//
				//				 = 1  +  sinh(t L') (m/L')  +  (cosh(t L') - 1) (m²/L'²)
				//			
				theLength = sqrt(-theLengthSquared);	//	here theLength is L' not L
				c1 = sinh(aTimeInterval * theLength) / theLength;
				c2 = (cosh(aTimeInterval * theLength) - 1.0) / (theLength * theLength);
			}
			else
			{
				c1 = aTimeInterval;
				c2 = aTimeInterval * aTimeInterval / 2.0;
			}
			
			break;

		default:	//	should never occur
			MatrixIdentity(m);
			c1 = 0.0;
			c2 = 0.0;
			break;
	}
	
	MatrixIdentity(theTerm0);

	MatrixCopy(theTerm1, m);
	ConstantTimesMatrix(c1, theTerm1, theTerm1);
	
	MatrixProduct(m, m, theTerm2);
	ConstantTimesMatrix(c2, theTerm2, theTerm2);
	
	MatrixCopy(aMotion, theTerm0);
	MatrixSum(aMotion, theTerm1, aMotion);
	MatrixSum(aMotion, theTerm2, aMotion);
}


void DifferentiatePosition(
	ktMatrix	aPosition,	//	input
	ktGeometry	aGeometry,	//	input
	ktVelocity	*aVelocity)	//	output
{
	//	Find aVelocity which, when integrated over unit time, yields aPosition.
	
	//	The unit time assumption lets us set t = 1 in all formulas
	//	from the comments in IntegrateVelocity().
	
	double	theTrace,
			theEpsilon,
			theFactor,
			vx2,
			vy2,
			vr2,
			theL,
			a,
			b;

	switch (aGeometry)
	{
		case GeometrySpherical:

			//	Our first task is to decide whether L is close to zero or not
			//	(see the comments in IntegrateVelocity() for a full discussion of L).
			//	In the general case,
			//
			//		tr(exp(m)) = tr(1) + sin(L) tr(m/L) + (1 - cos(L)) tr((m/L))
			//		           =   3   + sin(L)    0    + (1 - cos(L))    (-2)
			//		           =   1 + 2 cos(L)
			//
			//	implying that when L is small, tr(exp(m)) is close to 3,
			//	in which case tr(exp(m)) ~ 3 - L.  This agrees with the result
			//	obtained from the small-L formula for exp(m),
			//
			//		tr(exp(m)) = tr(1) + tr(m) + tr(m)/2
			//				   =   3   +  0    +  (-L)
			//
			//	Moreover, the result tr(exp(m)) = 1 + 2 cos(L) implies
			//	that exp(m) is a rotation through an angle L about some axis.
			//	In other words, L gives the final rotation angle directly!
			//
			//	In general the solution is not unique.  The axis of rotation
			//	for the infinitessimal motion must of course agree with the axis
			//	of rotation for aPosition, but it's up to us whether we choose
			//	to spin forwards or backwards, quickly or slowly, to reach aPosition.
			//	Without loss of generality we may assume the final rotation angle L
			//	lies in the range 0 <= L <= π.  We'll treat L ~ 0 and L ~ π
			//	as special cases.

			theTrace = aPosition[0][0] + aPosition[1][1] + aPosition[2][2];

			if (3.0 - theTrace < SMALL_L_LIMIT*SMALL_L_LIMIT)
			{
				//	L ~ 0, so we could read a first-order approximation to aVelocity
				//	directly from the off-diagonal entries of aPosition, but we can
				//	do even better and get a second-order approximation by averaging
				//	pairs of off-diagonal elements.
				aVelocity->x = (aPosition[2][0] - aPosition[0][2]) / 2.0;
				aVelocity->y = (aPosition[2][1] - aPosition[1][2]) / 2.0;
				aVelocity->r = (aPosition[0][1] - aPosition[1][0]) / 2.0;
			}
			else if (theTrace - (-1.0) < SMALL_L_LIMIT*SMALL_L_LIMIT)
			{
				//	L ~ π
				
				//	Don't ever let roundoff errors trigger a domain error.
				if (theTrace < -1.0)
					theTrace = -1.0;
				
				//	Let L = π - ε, and think in terms of ε.
				//
				//		tr(exp(m)) = 1 + 2 cos(L)
				//		           = 1 + 2 cos(π - ε)
				//		           = 1 - 2 cos(ε)
				//		           ~ -1 + ε
				//
				theEpsilon = sqrt(theTrace + 1.0);

				//	The Mathematica notebook MatrixExponentials.nb shows
				//	what the matrix looks like in this case.
				
				//	Differences of elements on the main diagonal give us
				//	the squares of vx, vy and vr.
				theFactor = (PI - theEpsilon)*(PI - theEpsilon)/(4.0 - theEpsilon*theEpsilon);
				vx2 = (-aPosition[0][0] + aPosition[1][1] - aPosition[2][2] + 1.0) * theFactor;
				vy2 = (+aPosition[0][0] - aPosition[1][1] - aPosition[2][2] + 1.0) * theFactor;
				vr2 = (-aPosition[0][0] - aPosition[1][1] + aPosition[2][2] + 1.0) * theFactor;

				//	Don't let roundoff error cause a domain error
				//	in the case that any of those quantities are zero.
				if (vx2 < 0.0)
					vx2 = 0.0;
				if (vy2 < 0.0)
					vy2 = 0.0;
				if (vr2 < 0.0)
					vr2 = 0.0;

				//	Unless ε is so small that it gets completely overwhelmed
				//	by roundoff errors, the differences between off-diagonal terms
				//	will give the correct signs of vx, vy and vr.
				//	(To handle the case where ε does indeed get overwhelmed
				//	by roundoff errors we could test which choice of signs
				//	best make the vector (-vy, vx, vr) a fixed point of aPosition,
				//	but it might not be worth writing the code for a situation that
				//	might never arise.  In the meantime, the worst that can happen
				//	is that the present code will return a half-turn about the wrong axis.)
				aVelocity->x = (aPosition[2][0] > aPosition[0][2] ? +sqrt(vx2) : -sqrt(vx2)) ;
				aVelocity->y = (aPosition[2][1] > aPosition[1][2] ? +sqrt(vy2) : -sqrt(vy2)) ;
				aVelocity->r = (aPosition[0][1] > aPosition[1][0] ? +sqrt(vr2) : -sqrt(vr2)) ;
			}
			else
			{
				//	In the generic case of L comfortably between 0 and π,
				//	we can read the velocity components as differences
				//	of off-diagonal terms of the matrix.
		
				theL		= acos((theTrace - 1.0)/2.0);
				theFactor	= theL / (2.0 * sin(theL));

				aVelocity->x = (aPosition[2][0] - aPosition[0][2]) * theFactor;
				aVelocity->y = (aPosition[2][1] - aPosition[1][2]) * theFactor;
				aVelocity->r = (aPosition[0][1] - aPosition[1][0]) * theFactor;
			}
			
			break;

		case GeometryEuclidean:

			//	In the Euclidean case, aVelocity represents the matrix
			//
			//			( 0   vr  0 )
			//		m = (-vr  0   0 )
			//			( vx  vy  0 )

			//	In the Euclidean case, with t = 1, the formula given
			//	in IntegrateVelocity() (see above) reduces to
			//
			//		         (  cos(vr)  sin(vr)  0  )
			//		exp(m) = ( -sin(vr)  cos(vr)  0  )
			//		         (    *        *      1  )
			//
			//	so we may read vr directly from the initial 2x2 block.
			//	It will be well defined up to a multiple of 2π.
			aVelocity->r = atan2(aPosition[0][1], aPosition[0][0]);
			
			//	If vr is zero (or close to zero)...
			if (fabs(aVelocity->r) < SMALL_L_LIMIT)
			{
				//	...then the * terms give vx and vy directly;
				aVelocity->x = aPosition[2][0];
				aVelocity->y = aPosition[2][1];
			}
			else  //	otherwise...
			{
				//	...it's a simple matter to solve for them.
				//	The Mathematica notebook MatrixExponentials.nb shows
				//	what the matrix looks like in this case (it's easy to replace
				//	the notebook's L = |vr| with the regular vr and check that
				//	all the signs come out right).
				a = (aVelocity->r*sin(aVelocity->r))/(2.0*(1.0 - cos(aVelocity->r)));
				b = aVelocity->r/2.0;
				aVelocity->x =  a*aPosition[2][0] + b*aPosition[2][1];
				aVelocity->y = -b*aPosition[2][0] + a*aPosition[2][1];
			}
			
			break;

		case GeometryHyperbolic:

			//	Begin as in the spherical case...
			//
			//		tr(exp(m)) = tr(1) + sin(L) tr(m/L) + (1 - cos(L)) tr((m/L))
			//		           =   3   + sin(L)    0    + (1 - cos(L))    (-2)
			//		           =   1 + 2 cos(L)
			//
			//	implying that when L is small, tr(exp(m)) is close to 3,
			//	in which case tr(exp(m)) ~ 3 - L.  This agrees with the result
			//	obtained from the small-L formula for exp(m),
			//
			//		tr(exp(m)) = tr(1) + tr(m) + tr(m)/2
			//				   =   3   +  0    +  (-L)
			//
			//	However, the hyperbolic case differs from the spherical case
			//	in that now L may be positive, zero or negative.
			//	When L is negative we may write L = i L' and interpret
			//	cos(L) as cosh(L').

			theTrace = aPosition[0][0] + aPosition[1][1] + aPosition[2][2];
			
			if (theTrace > 3.0 + SMALL_L_LIMIT*SMALL_L_LIMIT)
			{
				//	L < 0  (hyperbolic motion of the hyperbolic plane)
				
				//	Here the L is really L' in the above notation.
				theL		= acosh((theTrace - 1.0)/2.0);
				theFactor	= theL / (2.0 * sinh(theL));

				aVelocity->x = (aPosition[2][0] + aPosition[0][2]) * theFactor;
				aVelocity->y = (aPosition[2][1] + aPosition[1][2]) * theFactor;
				aVelocity->r = (aPosition[0][1] - aPosition[1][0]) * theFactor;
			}
			else if (theTrace < 3.0 - SMALL_L_LIMIT*SMALL_L_LIMIT)
			{
				//	L > 0  (elliptic motion of the hyperbolic plane)

				//	Don't ever let roundoff errors trigger a domain error.
				if (theTrace < -1.0)
					theTrace = -1.0;

				if (theTrace < -1.0 + SMALL_L_LIMIT*SMALL_L_LIMIT)
				{
					//	Deal with the special case that
					//	the rotation is approximately a half turn.
					//	Proceed as in the spherical case.
				
					//	Let L = π - ε, and think in terms of ε.
					//
					//		tr(exp(m)) = 1 + 2 cos(L)
					//		           = 1 + 2 cos(π - ε)
					//		           = 1 - 2 cos(ε)
					//		           ~ -1 + ε
					//
					theEpsilon = sqrt(theTrace + 1.0);

					//	The Mathematica notebook MatrixExponentials.nb shows
					//	what the matrix looks like in this case.
					
					//	Differences of elements on the main diagonal give us
					//	the squares of vx, vy and vr.
					theFactor = (PI - theEpsilon)*(PI - theEpsilon)/(4.0 - theEpsilon*theEpsilon);
					vx2 = (+aPosition[0][0] - aPosition[1][1] + aPosition[2][2] - 1.0) * theFactor;
					vy2 = (-aPosition[0][0] + aPosition[1][1] + aPosition[2][2] - 1.0) * theFactor;
					vr2 = (-aPosition[0][0] - aPosition[1][1] + aPosition[2][2] + 1.0) * theFactor;

					//	Don't let roundoff error cause a domain error
					//	in the case that any of those quantities are zero.
					if (vx2 < 0.0)
						vx2 = 0.0;
					if (vy2 < 0.0)
						vy2 = 0.0;
					if (vr2 < 0.0)
						vr2 = 0.0;

					//	Unless ε is so small that it gets completely overwhelmed
					//	by roundoff errors, the differences between off-diagonal terms
					//	will give the correct signs of vx, vy and vr.
					//	(To handle the case where ε does indeed get overwhelmed
					//	by roundoff errors we could test which choice of signs
					//	best make the vector (-vy, vx, vr) a fixed point of aPosition,
					//	but it might not be worth writing the code for a situation that
					//	might never arise.  In the meantime, the worst that can happen
					//	is that the present code will return a half-turn about the wrong axis.)
					aVelocity->x = (aPosition[2][0] + aPosition[0][2] > 0 ? +sqrt(vx2) : -sqrt(vx2)) ;
					aVelocity->y = (aPosition[2][1] + aPosition[1][2] > 0 ? +sqrt(vy2) : -sqrt(vy2)) ;
					aVelocity->r = (aPosition[0][1] - aPosition[1][0] > 0 ? +sqrt(vr2) : -sqrt(vr2)) ;
				}
				else
				{
					//	In the generic case of L comfortably between 0 and π,
					//	we can read the velocity components as differences
					//	of off-diagonal terms of the matrix.
			
					theL		= acos((theTrace - 1.0)/2.0);
					theFactor	= theL / (2.0 * sin(theL));

					aVelocity->x = (aPosition[2][0] + aPosition[0][2]) * theFactor;
					aVelocity->y = (aPosition[2][1] + aPosition[1][2]) * theFactor;
					aVelocity->r = (aPosition[0][1] - aPosition[1][0]) * theFactor;
				}
			}
			else
			{
				//	L ~ 0  (parabolic motion of the hyperbolic plane,
				//				or maybe a slight hyperbolic or elliptic motion)
				//
				//	The Mathematica notebook MatrixExponentials.nb shows
				//	what the small-L matrix looks like.
				//	Average pairs of off-diagonal elements.
				aVelocity->x = (aPosition[2][0] + aPosition[0][2]) / 2.0;
				aVelocity->y = (aPosition[2][1] + aPosition[1][2]) / 2.0;
				aVelocity->r = (aPosition[0][1] - aPosition[1][0]) / 2.0;
			}

			break;

		default:	//	should never occur
			break;
	}
}
