//	AVIMovie.c
//
//	The data structure
//
//		AVIMovie
//
//	along with the three functions
//
//		BeginAVIMovie()
//		EndAVIMovie()
//		WriteAVIMovieFrame()
//
//	comprise a self-contained system for writing AVI files.
//	For simplicity they assume the program is running in 32-bit color,
//	and expect data in ARGB format.  They need to know about
//	the Win32 API (including vfw.h) but know nothing about OpenGL.
//
//	The sample function MakeOpenGLMovie() illustrates the use
//	of the AVIMovie functions.  Unlike the latter, the former
//	relies on OpenGL.  A program would call it in place
//	of the usual rendering function, to write frames to a file.
//	Varying the contents of MakeOpenGLMovie() lets one
//	in effect script the movie.
//
//	KaleidoTile uses the symbol ENABLE_AVI_MOVIE
//	to bracket the menu command that calls MakeOpenGLMovie().
//	In addition, one must insure that the window's client
//	area's width and height are multiples of 4.
//
//	Jeff Weeks
//	June 2004
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt


////////////////////////////////////////////////////////////////
//
//		AVIMovie package
//
////////////////////////////////////////////////////////////////

#include "KaleidoTile-Win32.h"
#include <vfw.h>

typedef struct
{
	unsigned int		itsWidth,
						itsHeight,
						itsFrameRate,
						itsFrameCount;
	PAVIFILE			itsFile;
	PAVISTREAM			itsStream;
	AVICOMPRESSOPTIONS	itsCompressionOptions,
						*itsCompressionOptionsPtr;
	PAVISTREAM			itsCompressedStream;
} AVIMovie;

AVIMovie	*BeginAVIMovie(WCHAR *aFileName, unsigned int aWidth, unsigned int aHeight, unsigned int aFrameRate);
void		EndAVIMovie(AVIMovie **aMovie);
void		WriteAVIMovieFrame(AVIMovie *aMovie, void *somePixels);


AVIMovie *BeginAVIMovie(
	WCHAR			*aFileName,
	unsigned int	aWidth,
	unsigned int	aHeight,
	unsigned int	aFrameRate)
{
	ErrorText			theError	= NULL;
	AVIMovie			*theMovie	= NULL;
	AVISTREAMINFO		theInfo;
	BITMAPINFOHEADER	theFormat;

	//	aWidth and aHeight must both be multiples of 4.
	if (aWidth%4 || aHeight%4)
	{
		theError = "The frame's width and height much each be a multiple of 4.";
		goto CleanUpBeginAVIMovie;
	}

	theMovie = GET_MEMORY(sizeof(AVIMovie));
	if (theMovie == NULL)
	{
		theError = "Couldn't allocate memory for the AVIMovie struct.";
		goto CleanUpBeginAVIMovie;
	}

	theMovie->itsWidth		= aWidth;
	theMovie->itsHeight		= aHeight;
	theMovie->itsFrameRate	= aFrameRate;
	theMovie->itsFrameCount	= 0;

	//	Initialize various pointers to NULL
	//	to facilitate error cleanup.
	theMovie->itsFile					= NULL;
	theMovie->itsStream					= NULL;
	theMovie->itsCompressionOptionsPtr	= NULL;
	theMovie->itsCompressedStream		= NULL;

	AVIFileInit();

	if (AVIFileOpen(&theMovie->itsFile,
					aFileName,
					OF_CREATE | OF_WRITE,
					NULL) != AVIERR_OK)
	{
		theError = "AVIFileOpen() failed.";
		goto CleanUpBeginAVIMovie;
	}

	ZeroMemory(&theInfo, sizeof(AVISTREAMINFO));
	theInfo.fccType					= streamtypeVIDEO;
	theInfo.fccHandler				= 0;
	theInfo.dwScale					= 1;	//	always 1 for video
	theInfo.dwRate					= theMovie->itsFrameRate;
	theInfo.dwSuggestedBufferSize	= theMovie->itsWidth * theMovie->itsHeight * 4;
	SetRect(&theInfo.rcFrame, 0, 0, theMovie->itsWidth, theMovie->itsHeight);
	if (AVIFileCreateStream(theMovie->itsFile,
							&theMovie->itsStream,
							&theInfo) != AVIERR_OK)
	{
		theError = "AVIFileCreateStream() failed.";
		goto CleanUpBeginAVIMovie;
	}

	ZeroMemory(&theMovie->itsCompressionOptions, sizeof(AVICOMPRESSOPTIONS));
	theMovie->itsCompressionOptionsPtr = &theMovie->itsCompressionOptions;
	if (AVISaveOptions(NULL, 0, 1, &theMovie->itsStream, &theMovie->itsCompressionOptionsPtr) != true)
	{
		theMovie->itsCompressionOptionsPtr = NULL;	//	no resources will need to be freed
		theError = "AVISaveOptions() failed.";
		goto CleanUpBeginAVIMovie;
	}

	if (AVIMakeCompressedStream(&theMovie->itsCompressedStream,
								theMovie->itsStream,
								&theMovie->itsCompressionOptions,
								NULL) != AVIERR_OK)
	{
		theError = "AVIMakeCompressedStream() failed.";
		goto CleanUpBeginAVIMovie;
	}

	ZeroMemory(&theFormat, sizeof(BITMAPINFOHEADER));
	theFormat.biSize		= sizeof(BITMAPINFOHEADER);
	theFormat.biWidth		= theMovie->itsWidth;
	theFormat.biHeight		= theMovie->itsHeight;
	theFormat.biPlanes		= 1;
	theFormat.biBitCount	= 32;
	theFormat.biCompression	= BI_RGB;
	if (AVIStreamSetFormat(	theMovie->itsCompressedStream,
							0,
							&theFormat,
							sizeof(BITMAPINFOHEADER)) != AVIERR_OK)
	{
		theError = "AVIStreamSetFormat() failed.";
		goto CleanUpBeginAVIMovie;
	}

	return theMovie;

CleanUpBeginAVIMovie:
	ErrorMessage(theError, "BeginAVIMovie() Error");
	EndAVIMovie(&theMovie);
	return NULL;
}

void EndAVIMovie(AVIMovie **aMovie)
{
	if (*aMovie != NULL)
	{
		if ((*aMovie)->itsCompressedStream != NULL)
			AVIStreamRelease((*aMovie)->itsCompressedStream);

		if ((*aMovie)->itsCompressionOptionsPtr != NULL)
			AVISaveOptionsFree(1, &(*aMovie)->itsCompressionOptionsPtr);

		if ((*aMovie)->itsStream != NULL)
			AVIStreamRelease((*aMovie)->itsStream);

		if ((*aMovie)->itsFile != NULL)
			AVIFileRelease((*aMovie)->itsFile);

		AVIFileExit();

		FREE_MEMORY_SAFELY(*aMovie);
	}
}


void WriteAVIMovieFrame(
	AVIMovie	*aMovie,
	void		*somePixels)	//	32-bit argb format
{
	AVIStreamWrite(	aMovie->itsCompressedStream,
					aMovie->itsFrameCount,
					1,
					somePixels,
					aMovie->itsWidth * aMovie->itsHeight * 4,
					0,	//	AVIIF_KEYFRAME,
					NULL,
					NULL);
	aMovie->itsFrameCount++;
}

////////////////////////////////////////////////////////////////
//
//		sample MakeOpenGLMovie()
//
////////////////////////////////////////////////////////////////

#include <gl.h>		//	lives in gl/ on Win32

//	OpenGL 1.2 supports the following two definitions.
//	If OpenGL 1.2 is not present, glReadPixels() will fail.
#define GL_BGRA                           0x80E1
#define GL_UNSIGNED_INT_8_8_8_8_REV       0x8367


ErrorText MakeOpenGLMovie(WindowData *wd)
{
	ErrorText		theError;
	GLint			*thePixels;
	AVIMovie		*theMovie;
	RECT			theClientRect;
	unsigned int	theWidth,
					theHeight;

	static unsigned int	theFrameRate	= 24,
						theNumFrames	= 24;
	static WCHAR		*theTitle		= L"sample.avi";

	theError	= NULL;
	thePixels	= NULL;
	theMovie	= NULL;

	if (wd == NULL)
	{
		theError = "NULL argument";
		goto CleanUpMakeOpenGLMovie;
	}

	if (GetClientRect(wd->itsDrawingPanel, &theClientRect) == 0
	 || theClientRect.left != 0
	 || theClientRect.top  != 0)
	{
		theError = "Missing or invalid ClientRect";
		goto CleanUpMakeOpenGLMovie;
	}

	theWidth	= theClientRect.right;
	theHeight	= theClientRect.bottom;

	thePixels = GET_MEMORY(theWidth * theHeight * sizeof(GLint));
	if (thePixels == NULL)
	{
		theError = "Couldn't allocate thePixels.";
		goto CleanUpMakeOpenGLMovie;
	}

	theMovie = BeginAVIMovie(theTitle, theWidth, theHeight, theFrameRate);
	if (theMovie == NULL)
	{
		theError = "Failed call to BeginAVIMovie()";
		goto CleanUpMakeOpenGLMovie;
	}

	if ( ! wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext) )
	{
		theError = "Failed call to wglMakeCurrent()";
		goto CleanUpMakeOpenGLMovie;
	}

	//	Here's where you script the specific movie.
//INITIALIZE APPLICATION PARAMETERS IN PREPARATION FOR THE FIRST FRAME.
	while (theMovie->itsFrameCount < theNumFrames)
	{
		if (Render(&wd->md) != NULL)
		{
			theError = "Failed to render";
			goto CleanUpMakeOpenGLMovie;
		}

		glReadPixels(0, 0, theWidth, theHeight,
			GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, thePixels);

		WriteAVIMovieFrame(theMovie, thePixels);

//UPDATE APPLICATION PARAMETERS IN PREPARATION FOR THE NEXT FRAME.
	}


CleanUpMakeOpenGLMovie:

	wglMakeCurrent(NULL, NULL);

	if (theMovie != NULL)
		EndAVIMovie(&theMovie);

	FREE_MEMORY_SAFELY(thePixels);

	if (theError != NULL)
		ErrorMessage(theError, "MakeOpenGLMovie() Error");

	return theError;
}
