//	KaleidoTile-Win32-Utilities.c
//
//	Provides platform-dependent supporting functions
//	for the Win32 version of KaleidoTile.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Win32.h"
#include <string.h>		//	for wcslen() and wcscpy()
#include <mmsystem.h>	//	for sound functions


static bool		ConvertEndOfLineMarkers(WCHAR *aTextBuffer, unsigned int aTextBufferLength);
static DWORD	GetMaskShift(DWORD aMask);


#define CONVERSION_ERROR_STRING		L"<error converting text>"

WCHAR *GetLocalizedTextAsWCHAR(
	const char		*aKey,				//	zero-terminated UTF-8 string
	WCHAR			*aValueBuffer,		//	zero-terminated WCHAR string
	unsigned int	aValueBufferLength)	//	length of aValueBuffer in WCHARs (not bytes),
										//		including space for the terminating zero
{
	if (aValueBufferLength == 0)
	{
		ErrorMessage("GetLocalizedTextAsWCHAR() received zero-length output buffer.", "Internal Error");
		ExitProcess(4);
	}

	if (MultiByteToWideChar(
			CP_UTF8, MB_ERR_INVALID_CHARS,
			GetLocalizedText(aKey), -1,
			aValueBuffer, aValueBufferLength) == 0
	 || ! ConvertEndOfLineMarkers(aValueBuffer, aValueBufferLength) )
	{
		if (aValueBufferLength >= wcslen(CONVERSION_ERROR_STRING) + 1)
			wcscpy(aValueBuffer, CONVERSION_ERROR_STRING);
		else
		if (aValueBufferLength >= 2)
			wcscpy(aValueBuffer, L"?");
		else
			aValueBuffer[0] = 0;
	}

	return aValueBuffer;
}

static bool ConvertEndOfLineMarkers(
	WCHAR			*aTextBuffer,		//	zero-terminated WCHAR string
	unsigned int	aTextBufferLength)	//	length of aTextBuffer in WCHARs (not bytes),
										//		including space for the terminating zero;
										//		typically greater than the space occupied
										//		by the string itself
{
	//	Convert "\n" to "\r\n" if aValueBuffer is big enough to hold it.

	WCHAR	*r,	//	read  position
			*w;	//	write position

	//	Set r to the terminating zero.
	//	Set w a little bit further on, 
	//	to allow room to insert '\r' characters.
	r = w = aTextBuffer;
	while (*r != 0)
	{
		if (*r == L'\n')
			w++;
		r++;
		w++;
	}
	
	//	If we've got room to insert the '\r' characters...
	if (w < aTextBuffer + aTextBufferLength)
	{
		//	then read the string backwards...
		do
		{
			//	copying all characters...
			*w = *r;

			//	and inserting '\r' characters as necessary.
			if (*r == L'\n'
			 && --w >= aTextBuffer)	//	error check is logically unnecessary
			{
				*w = L'\r';
			}

		} while (--r >= aTextBuffer && --w >= aTextBuffer);
			
		return true;
	}
	else
		return false;
}


bool GetBasePath(
	char			*aPath,			//	output buffer, zero-terminated UTF-8 string
	unsigned int	aPathSize)		//	output buffer size, including space for terminating zero
{
	bool			theReturnValue	= false;
	WCHAR			*thePath		= NULL;	//	zero-terminated UTF-16 string
	unsigned int	theTerminator,
					i;

	//	Allocate a buffer for the UTF-16 version of the path.
	//	One UTF-16 character is guaranteed to be enough
	//	to handle any string that will ultimately fit
	//	into a UTF-8 path of aPathSize bytes.
	thePath = (WCHAR *) GET_MEMORY(aPathSize * sizeof(WCHAR));
	if (thePath == NULL)
		goto CleanUpGetBasePath;

	//	Get an absolute path for our executable file.
	//	GetModuleFileName() will return the length of the string,
	//	which we interpret as the position of the terminating zero.
	//	If the true path is too long for the buffer, presumably
	//	GetModuleFileName() allows room for the terminating zero,
	//	so assume a path that just barely fits has been truncated.
	//	If GetModuleFileName() fails for any other reason,
	//	it should return 0.
	theTerminator = GetModuleFileName(NULL, thePath, aPathSize);
	if (theTerminator == 0
	 || theTerminator >= aPathSize - 1)
	{
		goto CleanUpGetBasePath;
	}

	//	Remove the name of executable file itself,
	//	leaving only the path to the directory,
	//	including the trailing backslash.
	while (theTerminator > 0 && thePath[theTerminator - 1] != '\\')
		thePath[--theTerminator] = 0;
	
	//	Convert the path separator from '\' to '/'.
	for (i = 0; i < theTerminator; i++)
		if (thePath[i] == '\\')
			thePath[i] = '/';

	//	Convert to UTF-8.
	if (WideCharToMultiByte(CP_UTF8, 0, thePath, -1, aPath, aPathSize, NULL, NULL) == 0)
		goto CleanUpGetBasePath;

	//	Success!
	theReturnValue = true;

CleanUpGetBasePath:
	
	FREE_MEMORY_SAFELY(thePath);
	
	if ( ! theReturnValue )
		aPath[0] = 0;
	
	return theReturnValue;
}


ErrorText GetFileContents(
	const char		*aPathName,		//	absolute path as zero-terminated UTF-8 string
	unsigned int	*aNumRawBytes,	//	output, size of file's contents
	Byte			**someRawBytes)	//	output, copy of file's contents
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	ErrorText		theErrorMessage	= NULL;
	unsigned int	thePathLength	= 0;
	WCHAR			*thePathName	= NULL;
	HANDLE			theFile			= INVALID_HANDLE_VALUE;
	unsigned long	theNumBytesRead	= 0;

	//	Convert aPathName from UTF-8 to UTF-16.
	//	One UTF-16 character per UTF-8 character
	//	is guaranteed to suffice (even if the UTF-16 text
	//	includes surrogate pairs).
	thePathLength = 0;
	while (aPathName[thePathLength++] != 0)	//	counts the terminating zero as well
		;
	thePathName = (WCHAR *) GET_MEMORY(thePathLength * sizeof(WCHAR));
	if (thePathName == NULL)
	{
		theErrorMessage = "Couldn't get memory for UTF-16 path name in GetFileContents().";
		goto CleanUpGetFileContents;
	}
	if (MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aPathName, -1, thePathName, thePathLength) == 0)
	{
		theErrorMessage = "Couldn't convert path name from UTF-8 to UTF-16 in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Open the file.
	theFile = CreateFile(thePathName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (theFile == INVALID_HANDLE_VALUE)
	{
		theErrorMessage = "Couldn't open file in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Get the file size.
	*aNumRawBytes = GetFileSize(theFile, NULL);
	if (*aNumRawBytes == INVALID_FILE_SIZE)
	{
		theErrorMessage = "Couldn't get file size in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Allocate a buffer to read the file contents into.
	*someRawBytes = (Byte *) GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate memory for file's contents in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Read the file contents into the buffer.
	if (ReadFile(theFile, *someRawBytes, *aNumRawBytes, &theNumBytesRead, NULL) == 0
	 || theNumBytesRead != *aNumRawBytes)
	{
		theErrorMessage = "Couldn't read file's contents in GetFileContents().";
		goto CleanUpGetFileContents;
	}

CleanUpGetFileContents:
	
	if (theFile != INVALID_HANDLE_VALUE)
		CloseHandle(theFile);

	FREE_MEMORY_SAFELY(thePathName);
	
	if (theErrorMessage != NULL)
	{
		*aNumRawBytes = 0;
		FREE_MEMORY_SAFELY(*someRawBytes);
	}

	return theErrorMessage;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


void SetAlphaTextureFromString(
	unsigned int	aTextureName,
	const char		*aString,		//	null-terminated UTF-8 string
	unsigned int	aWidth,
	unsigned int	aHeight,
	char			*aFontName,		//	null-terminated ASCII string
	unsigned int	aFontSize,		//	height in pixels, excluding descent
	unsigned int	aFontDescent,	//	vertical space below baseline, in pixels
	bool			aCenteringFlag,	//	Center the text horizontally in the texture image?
	unsigned int	aMargin,		//	If aCenteringFlag == false, what horizontal inset to use?
	bool			aShaderFlag,
	ErrorText		*aFirstError)
{
	ErrorText		theErrorMessage		= NULL;
	HDC				theScreenDC			= NULL,
					theOffscreenDC		= NULL;
	HBITMAP			theBitmap			= NULL;
	TEXTMETRIC		theTextMetrics;
	LOGFONT			theLogicalFont		= {0, 0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, 0, L""};
	HFONT			theFont				= NULL;
	WCHAR			theString[128]		= L"";
	unsigned int	theStringLength		= 0;	//	theStringLength will ultimately include the terminating zero
	Byte			*theRGBABytes		= NULL;
	BITMAPINFO		theBitmapInfo		=	{
												{sizeof(BITMAPINFOHEADER), 0, 0, 1, 32, BI_RGB, 0, 0, 0, 0, 0},
												{{0,0,0,0}}
											};
	Byte			*theAlphaBytes		= NULL;
	unsigned int	i;

	//	Assumes desired OpenGL context is already active.

	if ( ! IsPowerOfTwo(aWidth)
	 ||  ! IsPowerOfTwo(aHeight) )
	{
		theErrorMessage = "aWidth and aHeight must be powers of two.";
		goto CleanUpSetAlphaTextureFromString;
	}

	theScreenDC = CreateDC(L"DISPLAY", NULL, NULL, NULL);
	//	Don't panic if theScreenDC == NULL.
	//	Even without it we can still get a default monochrome bitmap.

	theOffscreenDC = CreateCompatibleDC(theScreenDC);	//	OK even if theScreenDC == NULL
	if (theOffscreenDC == NULL)
	{
		theErrorMessage = "Couldn't create device context.";
		goto CleanUpSetAlphaTextureFromString;
	}

	//	Unfortunately SelectObject() will insist on either a monochrome bitmap
	//	or a bitmap with the same color organization as the device
	//	(see Petzold page 666), so we can't use an obvious greyscale bitmap.
	//	Normally we'll use a color bitmap (for good antialiasing), but
	//	if for any reason theScreenDC == NULL we'll settle for monochrome
	//	(no antialiasing) instead.
	if (theScreenDC != NULL)
		theBitmap = CreateCompatibleBitmap(theScreenDC,    aWidth, aHeight);	//	same pixel format as screen
	else
		theBitmap = CreateCompatibleBitmap(theOffscreenDC, aWidth, aHeight);	//	monochrome
	if (theBitmap == NULL)
	{
		theErrorMessage = "Couldn't create bitmap.";
		goto CleanUpSetAlphaTextureFromString;
	}
	SelectObject(theOffscreenDC, theBitmap);

	SelectObject(theOffscreenDC, (HBRUSH)GetStockObject(BLACK_BRUSH));	//	makes background transparent

	theLogicalFont.lfHeight	= -aFontSize;	//	negative value indicates character height minus internal leading
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aFontName, -1, theLogicalFont.lfFaceName, BUFFER_LENGTH(theLogicalFont.lfFaceName));
	theFont = CreateFontIndirect(&theLogicalFont);
	SelectObject(theOffscreenDC, theFont);

	GetTextMetrics(theOffscreenDC, &theTextMetrics);

	SetTextColor(theOffscreenDC, 0x00FFFFFF);
	SetBkMode(theOffscreenDC, TRANSPARENT);
	
	//	For testing purposes only.
//	FillRect(theOffscreenDC, &(RECT){0,0,aWidth,aHeight}, (HBRUSH)GetStockObject(GRAY_BRUSH));

	theStringLength = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aString, -1, theString, BUFFER_LENGTH(theString));
	if (theStringLength == 0)
	{
		theErrorMessage = "Input string too long for buffer.";
		goto CleanUpSetAlphaTextureFromString;
	}

	//	Let's set the text alignment to TA_BOTTOM (rather than TA_BASELINE)
	//	and ignore aFontDescent.
	UNUSED_PARAMETER(aFontDescent);
	if (aCenteringFlag)
	{
		SetTextAlign(theOffscreenDC, TA_CENTER | TA_BOTTOM);
		TextOutW(theOffscreenDC, aWidth/2, aHeight, theString, theStringLength - 1);
	}
	else
	{
		SetTextAlign(theOffscreenDC, TA_LEFT | TA_BOTTOM);
		TextOutW(theOffscreenDC, aMargin, aHeight, theString, theStringLength - 1);
	}

	theRGBABytes = (Byte *) GET_MEMORY(aWidth*aHeight*4);
	if (theRGBABytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theRGBABytes.";
		goto CleanUpSetAlphaTextureFromString;
	}

	theBitmapInfo.bmiHeader.biWidth  = aWidth;
	theBitmapInfo.bmiHeader.biHeight = aHeight;
	if (aHeight != (unsigned int) GetDIBits(theOffscreenDC, theBitmap, 0, aHeight, theRGBABytes, &theBitmapInfo, DIB_RGB_COLORS))
	{
		theErrorMessage = "Couldn't read bitmap bits.";
		goto CleanUpSetAlphaTextureFromString;
	}

	theAlphaBytes = (Byte *) GET_MEMORY(aWidth*aHeight*1);
	if (theAlphaBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theAlphaBytes.";
		goto CleanUpSetAlphaTextureFromString;
	}

	//	Use the first component (red? blue?) of each pixel in theBitmap
	//	as the alpha value for the texture.
	for (i = 0; i < aWidth*aHeight; i++)
		theAlphaBytes[i] = theRGBABytes[4*i + 0];

	//	Convert the raw bitmap image to an OpenGL alpha-only texture.
	SetTextureImageFancy(aTextureName, aWidth, aHeight, 1, theAlphaBytes, aShaderFlag);

CleanUpSetAlphaTextureFromString:

	FREE_MEMORY_SAFELY(theAlphaBytes);

	FREE_MEMORY_SAFELY(theRGBABytes);

	if (theScreenDC != NULL)
		DeleteDC(theScreenDC);

	if (theOffscreenDC != NULL)		//	Delete theOffscreenDC before deleting objects selected into it.
		DeleteDC(theOffscreenDC);
	
	if (theFont != NULL)
		DeleteObject(theFont);

	if (theBitmap != NULL)
		DeleteObject(theBitmap);

	if (theErrorMessage != NULL)
	{
		//	Substitute a transparent texture for the missing one.
		SetTextureImageFancy(aTextureName, 1, 1, 1, (Byte [1]){0x00}, aShaderFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}


bool GetImageToPaste(
	unsigned int	*aWidth,
	unsigned int	*aHeight,
	PixelRGBA		**aPixelArray)
{
	ErrorText			theErrorMessage		= NULL;
	bool				theClipboardIsOpen	= false;
	HANDLE				theClipboardHandle	= NULL;
	void				*theClipboardImage	= NULL;
	BITMAPINFOHEADER	*theHeader			= NULL;
	unsigned int		theColorTableSize	= 0,
						i,
						j;
	RGBQUAD				*theColorTable		= NULL;
	Byte				*theBits			= NULL,
						*theSrcByte			= NULL;
	PixelRGBA			*theDstPixel		= NULL;
	DWORD				theMask[3],
						theShift[3],
						theSrcDepth,
						theSrcStride;
	Byte				theIndex,
						r,
						g,
						b;
	WORD				theSrcWord;
	DWORD				theSrcDword;

	//	Test for uninitialized inputs (but don't go to CleanUpGetImageToPaste,
	//	because we don't want to try to free what could be an invalid pointer).
	if ( *aPixelArray != NULL || *aWidth != 0 || *aHeight != 0 )
		return false;	//	should never occur

	if (OpenClipboard(NULL) == 0)
	{
		theErrorMessage = "Couldn't open clipboard.";
		goto CleanUpGetImageToPaste;
	}
	theClipboardIsOpen = true;

	//	Note:  GetClipboardData(CF_DIB) is guaranteed to give
	//	a plain BITMAPINFOHEADER followed by the required data.
	//	It never gives a BITMAPV4HEADER or BITMAPV5HEADER.
	theClipboardHandle = GetClipboardData(CF_DIB);
	if (theClipboardHandle == NULL)
	{
		//	Fail silently.  The caller will report the error.
		//	(Note:  The code at CleanUpGetImageToPaste treats
		//	an empty error message as a special case.)
		theErrorMessage = "";	//	"No bitmap found on clipboard.";
		goto CleanUpGetImageToPaste;
	}
	theClipboardImage = GlobalLock(theClipboardHandle);
	if (theClipboardImage == NULL)
	{
		theErrorMessage = "Couldn't lock clipboard data.";
		goto CleanUpGetImageToPaste;
	}

	theHeader = (BITMAPINFOHEADER *) theClipboardImage;

	//	Negative height means image isn't upside-down.
	if (theHeader->biHeight < 0)
	{
		//	...but let's assume for now we won't encounter that convention.
		theErrorMessage = "Sorry, this program can't handle the \"negative bitmap height\" convention.";
		goto CleanUpGetImageToPaste;
	}
	if (theHeader->biWidth < 0)
	{
		theErrorMessage = "Clipboard image has negative width.";
		goto CleanUpGetImageToPaste;
	}
	
	//	Record the (non-negative) height and width.
	*aWidth  = theHeader->biWidth;
	*aHeight = theHeader->biHeight;

	//	Complain if image is empty.
	//	(Later code relies on positive biWidth and biHeight.)
	if (*aWidth  == 0
	 || *aHeight == 0)
	{
		theErrorMessage = "Clipboard image has zero width and/or zero height";
		goto CleanUpGetImageToPaste;
	}

	//	How big is the color table, if any?
	switch (theHeader->biBitCount)
	{
		case 8:
			theColorTableSize = theHeader->biClrUsed;
			//	For 8-bit color, by convention, theHeader->biClrUsed = 0
			//	implies the maximum value, in this case 256.
			if (theColorTableSize == 0)
				theColorTableSize = 256;
			break;

		case 16:
		case 24:
		case 32:
			//	24-bit DIBs don't really need a color table,
			//	but in theory they can include one as a "hint"
			//	for mapping the image to a paletted display.
			//	16- and 32-bit DIBs may, depending on their format,
			//	use the color table to store bit-layout masks.
			theColorTableSize = theHeader->biClrUsed;
			break;

		default:
			theErrorMessage = "Bitmap must be in 8-bit, 16-bit, 24-bit or 32-bit format.";
			goto CleanUpGetImageToPaste;
	}

	//	Locate the color table (if any) and the bits themselves.
	theColorTable	= (RGBQUAD *)(
					  ((Byte *) theClipboardImage)
					+ theHeader->biSize );
	theBits			= ((Byte *) theClipboardImage)
					+ theHeader->biSize
					+ theColorTableSize * sizeof(RGBQUAD);

	//	Allocate memory for the new image.
	*aPixelArray = (PixelRGBA *) GET_MEMORY(*aWidth * *aHeight * sizeof(PixelRGBA));
	if (*aPixelArray == NULL)
	{
		theErrorMessage = "Couldn't allocate memory for bitmap.";
		goto CleanUpGetImageToPaste;
	}

	//	Precompute masks and shifts for 16- and 32-bit color formats.
	if (theHeader->biBitCount == 16
	 || theHeader->biBitCount == 32)
	{
		switch (theHeader->biCompression)
		{
			case BI_RGB:
				if (theHeader->biBitCount == 16)
				{
					//	1-5-5-5 xrgb color
					theMask[0]	= 0x00007C00;	// red
					theShift[0]	= 10;
					theMask[1]	= 0x000003E0;	// green
					theShift[1]	=  5;
					theMask[2]	= 0x0000001F;	// blue
					theShift[2]	=  0;
				}
				else	//	theHeader->biBitCount == 32
				{
					//	8-8-8-8 argb color
					theMask[0]	= 0x00FF0000;	// red
					theShift[0]	= 16;
					theMask[1]	= 0x0000FF00;	// green
					theShift[1]	=  8;
					theMask[2]	= 0x000000FF;	// blue
					theShift[2]	=  0;
				}
				break;

			case BI_BITFIELDS:
				//	custom rgb color
				theMask[0]	= *(DWORD *)&theColorTable[0];	// red
				theShift[0]	= GetMaskShift(theMask[0]);
				theMask[1]	= *(DWORD *)&theColorTable[1];	// green
				theShift[1]	= GetMaskShift(theMask[1]);
				theMask[2]	= *(DWORD *)&theColorTable[2];	// blue
				theShift[2]	= GetMaskShift(theMask[2]);
				break;

			default:	//	should not occur
				theMask[0]	= 0x00000000;	// red
				theShift[0]	= 0;
				theMask[1]	= 0x00000000;	// green
				theShift[1]	= 0;
				theMask[2]	= 0x00000000;	// blue
				theShift[2]	= 0;
				break;
		}
	}

	//	Copy the image.

	theSrcDepth  = theHeader->biBitCount / 8;	//	assumes 8-, 16-, 24- or 32-bit image
	theSrcStride = *aWidth * theSrcDepth;
	while (theSrcStride % 4 != 0)
		theSrcStride++;

	for (i = 0; i < *aHeight; i++)
	{
		theSrcByte  = &theBits[i*theSrcStride];
		theDstPixel = &(*aPixelArray)[i*(*aWidth)];

		for (j = 0; j < *aWidth; j++)
		{
			//	Read the RGB values of the source pixel.
			switch (theHeader->biBitCount)
			{
				case 8:
					//	Read an indexed color.
					theIndex = *theSrcByte;
					r = theColorTable[theIndex].rgbRed;
					g = theColorTable[theIndex].rgbGreen;
					b = theColorTable[theIndex].rgbBlue;
					break;

				case 16:
					//	Read a color using pre-computed masks and shifts.
					theSrcWord = * (WORD *) theSrcByte;
					r = (theSrcWord & theMask[0]) >> theShift[0];	// red
					g = (theSrcWord & theMask[1]) >> theShift[1];	// green
					b = (theSrcWord & theMask[2]) >> theShift[2];	// blue
					break;

				case 24:
					//	Read a bgr color.
					b = theSrcByte[0];	// blue
					g = theSrcByte[1];	// green
					r = theSrcByte[2];	// red
					break;

				case 32:
					//	Read a color using pre-computed masks and shifts.
					theSrcDword = * (DWORD *) theSrcByte;
					r = (theSrcDword & theMask[0]) >> theShift[0];	// red
					g = (theSrcDword & theMask[1]) >> theShift[1];	// green
					b = (theSrcDword & theMask[2]) >> theShift[2];	// blue
					break;

				default:	//	should not occur -- already excluded in preceding code
					r = g = b = 0xFF;
					break;
			}
			theSrcByte += theSrcDepth;

			//	Write the RGB values to the destination pixel.
			theDstPixel->r = r;
			theDstPixel->g = g;
			theDstPixel->b = b;
			theDstPixel->a = 0xFF;	//	fully opaque
			theDstPixel++;
		}
	}

CleanUpGetImageToPaste:

	if (theClipboardHandle != NULL
	 && theClipboardImage  != NULL)
		GlobalUnlock(theClipboardHandle);

	if (theClipboardIsOpen)
		CloseClipboard();

	if (theErrorMessage != NULL)
	{
		*aWidth  = 0;
		*aHeight = 0;
		FREE_MEMORY_SAFELY(*aPixelArray);

		//	Special case:  If the clipboard is empty,
		//	the code above will pass an empty error message ""
		//	as a signal that we should return false
		//	without displaying any error message.

		if (theErrorMessage[0] != 0)
			ErrorMessage(theErrorMessage, "Internal Error");

		return false;
	}
	else
		return true;
}


static DWORD GetMaskShift(DWORD aMask)
{
	DWORD	aShift;

	aShift = 0;

	if (aMask != 0)
	{
		while ( ! (aMask & 0x00000001))
		{
			aShift++;
			aMask >>= 1;
		}
	}

	return aShift;
}


void FreeImageToPaste(PixelRGBA **aPixelArray)
{
	FREE_MEMORY_SAFELY(*aPixelArray);
}


void PlayWAV(
	char	*aRelativePath)	//	zero-terminated UTF-8 string with path separator '/'
{
	ErrorText	theErrorMessage	= NULL;
	char		theAbsolutePathUTF8[2048];
	WCHAR		theAbsolutePathUTF16[2048];

	static bool	theErrorMessageWasGiven = false;

	if (GetAbsolutePath("Sounds", aRelativePath, theAbsolutePathUTF8, BUFFER_LENGTH(theAbsolutePathUTF8)) != NULL)
	{
		theErrorMessage = "Couldn't get absolute path for sound file.  (Path length might be too long.)";
		goto CleanUpPlayWAV;
	}

	if (MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, theAbsolutePathUTF8, -1, theAbsolutePathUTF16, BUFFER_LENGTH(theAbsolutePathUTF16)) == 0)
	{
		theErrorMessage = "Couldn't convert path name from UTF-8 to UTF-16.";
		goto CleanUpPlayWAV;
	}

	if ( ! PlaySound(theAbsolutePathUTF16, NULL, SND_FILENAME | SND_ASYNC | SND_NOWAIT) )
	{
		theErrorMessage = "Couldn't play the requested sound.  (One possible problem is that WinXP imposes a 256-character maximum on the file's path name.)";
		goto CleanUpPlayWAV;
	}

CleanUpPlayWAV:
	
	//	If an error occurs, report it, but only once per session.
	if (theErrorMessage != NULL
	 && ! theErrorMessageWasGiven )
	{
		ErrorMessage(theErrorMessage, "Error in function PlayWAV()");
		theErrorMessageWasGiven = true;
	}
}


bool GetUserPrefBool(
	char	*aKey)	//	ASCII
{
	HKEY	theRegistryKey;
	LONG	theQueryError;
	DWORD	theValue,
			theNumBytes;

	//	Try to read a previously saved user pref
	//	from the Windows registry.
	if (RegOpenKeyEx(HKEY_CURRENT_USER, TEXT("Software\\Geometry Games\\KaleidoTile"), 0, KEY_READ, &theRegistryKey) == ERROR_SUCCESS)
	{
		theNumBytes = sizeof(DWORD);
		theQueryError = RegQueryValueExA(theRegistryKey, aKey, NULL, NULL, (BYTE *)&theValue, &theNumBytes);
		RegCloseKey(theRegistryKey);
		if (theQueryError == ERROR_SUCCESS && theNumBytes == sizeof(DWORD))
			return (bool)theValue;
	}

	//	No previously saved user pref was found.
	//	Return a reasonable default value.
	if (strcmp(aKey, "multisampling"			) == 0)
		return true;
	if (strcmp(aKey, "shaders"					) == 0)
		return true;
	if (strcmp(aKey, "fallback message given"	) == 0)
		return false;
	return false;
}

void SetUserPrefBool(
	char	*aKey,	//	ASCII
	bool	aValue)
{
	HKEY	theRegistryKey;
	DWORD	theValue;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, TEXT("Software\\Geometry Games\\KaleidoTile"),
						0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &theRegistryKey, NULL)
		== ERROR_SUCCESS)
	{
		theValue = (DWORD)aValue;
		RegSetValueExA(theRegistryKey, aKey, 0, REG_DWORD, (BYTE *)&theValue, sizeof(DWORD));
		RegCloseKey(theRegistryKey);
	}
}


void ErrorMessage(
	ErrorText	aMessage,	//	UTF-8
	ErrorText	aTitle)		//	UTF-8
{
	//	Translate the incoming messages from UTF-8 to UTF-16
	//	and display them with MessageBoxW().

	bool			theMessageWasGiven	= false;
	unsigned int	theMessageLength	= 0,	//	number of unicode characters including terminating zero
					theTitleLength		= 0;
	WCHAR			*theMessageBuffer	= NULL,	//	buffer for UTF-16 translation including terminating zero
					*theTitleBuffer		= NULL;
	DWORD			theTextDirection;

	theMessageLength	= MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aMessage, -1, NULL, 0);
	if (theMessageLength == 0)		//	MultiByteToWideChar() failed, possibly because of bad input
		goto CleanUpErrorMessage;
	theMessageBuffer	= GET_MEMORY(theMessageLength * sizeof(WCHAR));
	if (theMessageBuffer == NULL)	//	deep trouble
		goto CleanUpErrorMessage;
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aMessage, -1, theMessageBuffer, theMessageLength);

	theTitleLength		= MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aTitle, -1, NULL, 0);
	if (theTitleLength == 0)		//	MultiByteToWideChar() failed, possibly because of bad input
		goto CleanUpErrorMessage;
	theTitleBuffer		= GET_MEMORY(theTitleLength * sizeof(WCHAR));
	if (theTitleBuffer == NULL)		//	deep trouble
		goto CleanUpErrorMessage;
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aTitle, -1, theTitleBuffer, theTitleLength);

	theTextDirection = LanguageReadsRightToLeft() ? (MB_RTLREADING | MB_RIGHT) : 0;

	MessageBoxW(NULL, theMessageBuffer, theTitleBuffer, MB_OK | MB_TASKMODAL | theTextDirection);

	theMessageWasGiven = true;

CleanUpErrorMessage:

	FREE_MEMORY_SAFELY(theMessageBuffer);
	FREE_MEMORY_SAFELY(theTitleBuffer  );

	//	If something went wrong...
	if ( ! theMessageWasGiven )
	{
		//	report the present failure as well as...
		MessageBoxA(NULL,
					"Original error message to follow...",
					"Error in ErrorMessage()",
					MB_OK | MB_TASKMODAL | MB_ICONERROR);

		//	the original error message.  The UTF-8 will be interpreted
		//	relative to the local code page, but at least any 7-bit characters
		//	should display OK.
		MessageBoxA(NULL,
					aMessage,
					aTitle,
					MB_OK | MB_TASKMODAL | MB_ICONERROR);
	}
}


bool LanguageReadsRightToLeft(void)
{
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ar")		//	Arabic
	 || SAME_LANGUAGE_CODE(gLanguageCode, "fa"))	//	Persian
		return true;
	else
		return false;
}
